"""CleanData: Example Class that demonstrates data cleanup for Light DataSources using Pandas"""

# Local imports
from sageworks.transforms.transform import Transform, TransformInput, TransformOutput
from sageworks.transforms.pandas_transforms.data_to_pandas import DataToPandas
from sageworks.transforms.pandas_transforms.pandas_to_data import PandasToData


class CleanData(Transform):
    def __init__(self, input_uuid: str = None, output_uuid: str = None):
        """CleanData: Class for filtering, sub-setting, and value constraints on Light DataSources uses Pandas"""

        # Call superclass init
        super().__init__(input_uuid, output_uuid)

        # Set up all my instance attributes
        self.input_type = TransformInput.DATA_SOURCE
        self.output_type = TransformOutput.DATA_SOURCE

    def transform(self):
        """Pull the input DataSource, clean the data, and output to a DataSource"""

        """
        Notes for later:
        Cleaning data typically involves two phases: Identification and Remediation.

        - Identification
            - Look for NaNs/NULL
            - Make sure all datatypes are correct (df.info()) if not .. fix...
            - Look at distributions/histograms
            - Look for outliers
        - Remediation
            - Drop
            - Fill/Replace
            - Impute the value (using inference/context to fill in a value)
        """

        # Grab the Input (Data Source)
        input_df = DataToPandas(self.input_uuid).get_output()  # Shorthand for transform, get_output

        # Drop Rows that have ANY NaNs in them
        orig_rows = len(input_df)
        input_df = input_df.dropna(axis=0, how='any')
        if len(input_df) != orig_rows:
            self.log.info(f"Dropping {orig_rows - len(input_df)} rows that have a NaN in them")

        # Drop Columns that have ANY NaNs in them
        orig_columns = input_df.columns.tolist()
        input_df = input_df.dropna(axis=1, how='any')
        remaining_columns = input_df.columns.tolist()
        if remaining_columns != orig_columns:
            dropped_columns = list(set(remaining_columns).difference(set(orig_columns)))
            self.log.info(f"Dropping {dropped_columns} columns that have a NaN in them")

        # Now publish to the output location
        output_data = PandasToData(self.output_uuid)
        output_data.set_input(input_df)
        output_data.transform()


# Simple test of the CleanData functionality
def test():
    """Test the CleanData Class"""
    import pandas as pd
    from sageworks.artifacts.data_sources.athena_source import AthenaSource

    # Create the class with inputs and outputs and invoke the transform
    input_uuid = 'aqsol_data'
    output_uuid = 'aqsol_data_clean'
    CleanData(input_uuid, output_uuid).transform()

    # Grab the output and query it for a dataframe
    output = AthenaSource(output_uuid)
    df = output.query(f"select * from {output_uuid} limit 5")

    # Setup Pandas output options
    pd.set_option('display.max_colwidth', 15)
    pd.set_option('display.max_columns', 15)
    pd.set_option('display.width', 1000)

    # Show the dataframe
    print(df)


if __name__ == "__main__":
    test()



"""Connector: Abstract Base Class for pulling/refreshing AWS Service metadata"""
from abc import ABC, abstractmethod

import logging

# SageWorks Imports
from sageworks.utils.logging import logging_setup

# Set up logging
logging_setup()


class Connector(ABC):
    def __init__(self):
        """Connector: Abstract Base Class for pulling/refreshing AWS Service metadata"""
        self.log = logging.getLogger(__name__)

    @abstractmethod
    def check(self) -> bool:
        """Can we connect to this service?"""
        pass

    @abstractmethod
    def refresh(self) -> bool:
        """Refresh data/metadata associated with this service"""
        pass

    @abstractmethod
    def metadata(self) -> dict:
        """Return all the metadata for this service"""
        pass



"""FeatureStore: Helper Class for the AWS Feature Store Service"""
import sys
import argparse
import awswrangler as wr
import json
from sagemaker.session import Session

# SageWorks Imports
from sageworks.aws_service_broker.aws_service_connectors.connector import Connector


class FeatureStore(Connector):
    def __init__(self):
        """"FeatureStore: Helper Class for the AWS Feature Store Service"""
        # Call SuperClass Initialization
        super().__init__()

        # Set up our SageMaker Session and SageMaker Client
        self.sm_session = Session()
        self.sm_client = self.sm_session.boto_session.client("sagemaker")

        # Set up our internal data storage
        self.feature_data = {}

    def check(self) -> bool:
        """Check if we can reach/connect to this AWS Service"""
        try:
            self.sm_client.list_feature_groups()
            return True
        except Exception as e:
            self.log.critical(f"Error connecting to AWS Feature Store Service: {e}")
            return False

    def refresh(self):
        """Load/reload the tables in the database"""
        # Grab all the Feature Groups in the AWS Feature Store
        print("Reading Feature Store Database...")
        _feature_groups = self.sm_client.list_feature_groups()['FeatureGroupSummaries']
        _fg_names = [feature_group['FeatureGroupName'] for feature_group in _feature_groups]

        # Get the details for each Feature Group and convert to a data structure with direct lookup
        self.feature_data = {name: self._feature_group_details(name) for name in _fg_names}

        # Also add additional details under the sageworks section for each Feature Group
        for fg_name in _fg_names:
            add_data = {'athena_database': self.athena_database_name(fg_name),
                        'athena_table': self.athena_table_name(fg_name),
                        's3_storage': self.s3_storage(fg_name)}
            self.feature_data[fg_name]['sageworks'] = add_data

    def metadata(self) -> dict:
        """Get all the table information in this database"""
        return self.feature_data

    def feature_group_names(self) -> list:
        """Get all the feature group names in this database"""
        return list(self.feature_data.keys())

    def feature_group_details(self, feature_group_name: str) -> dict:
        """Get the details for a specific feature group"""
        return self.feature_data.get(feature_group_name)

    def athena_database_name(self, feature_group_name: str) -> str:
        """Get the Athena Database Name for a specific feature group"""
        return self.feature_data[feature_group_name]['OfflineStoreConfig']['DataCatalogConfig']['Database']

    def athena_table_name(self, feature_group_name: str) -> str:
        """Get the Athena Table Name for a specific feature group"""
        return self.feature_data[feature_group_name]['OfflineStoreConfig']['DataCatalogConfig']['TableName']

    def s3_storage(self, feature_group_name: str) -> str:
        """Get the S3 Location for a specific feature group"""
        return self.feature_data[feature_group_name]['OfflineStoreConfig']['S3StorageConfig']['ResolvedOutputS3Uri']

    def get_feature_group_tags(self, feature_group_name: str) -> list:
        """Get the table tag list for the given table name"""
        feature_group = self.feature_group_details(feature_group_name)
        return json.loads(feature_group['Parameters'].get('tags', '[]'))

    def set_feature_group_tags(self, database: str, table_name: str, tags: list):
        """Set the tags for a specific feature group"""
        wr.catalog.upsert_table_parameters(parameters={'tags': json.dumps(tags)},
                                           database=database,
                                           table=table_name)

    def add_feature_group_tags(self, database: str,  table_name: str, tags: list):
        """Add some the tags for a specific feature set"""
        current_tags = json.loads(wr.catalog.get_table_parameters(database, table_name).get('tags'))
        new_tags = list(set(current_tags).union(set(tags)))
        wr.catalog.upsert_table_parameters(parameters={'tags': json.dumps(new_tags)},
                                           database=database,
                                           table=table_name)

    def _feature_group_details(self, feature_group_name: str) -> dict:
        """Internal: Do not call this method directly, use feature_group_details() instead"""

        # Grab the Feature Group details from the AWS Feature Store
        details = self.sm_client.describe_feature_group(FeatureGroupName=feature_group_name)
        return details

    def snapshot_query(self, feature_group_name: str) -> str:
        """Construct an Athena 'snapshot' query for the given feature group"""
        database = self.athena_database_name(feature_group_name)
        table = self.athena_table_name(feature_group_name)
        event_time = 'EventTime'
        record_id = 'id'
        query = f"""
            SELECT *
            FROM
                (SELECT *,
                     row_number()
                    OVER (PARTITION BY {record_id}
                ORDER BY  {event_time} desc, Api_Invocation_Time DESC, write_time DESC) AS row_num
                FROM {database}.{table})
            WHERE row_num = 1 and
            NOT is_deleted;
        """
        return query


if __name__ == '__main__':
    from pprint import pprint

    # Collect args from the command line
    parser = argparse.ArgumentParser()
    args, commands = parser.parse_known_args()

    # Check for unknown args
    if commands:
        print('Unrecognized args: %s' % commands)
        sys.exit(1)

    # Create the class and get the AWS Feature Store details
    feature_store = FeatureStore()
    feature_store.refresh()

    # List the Feature Groups
    print('Feature Groups:')
    for fname in feature_store.feature_group_names():
        print(f"\t{fname}")

    # Get the details for a specific Feature Group
    my_group = fname
    group_info = feature_store.feature_group_details(my_group)
    pprint(group_info)

    # Get the Athena Database Name for this Feature Group
    my_database = feature_store.athena_database_name(my_group)

    # Get the Athena Table Name for this Feature Group
    my_table = feature_store.athena_table_name(my_group)

    # Get the Athena Query for this Feature Group
    my_query = feature_store.snapshot_query(my_group)

    """ TAGS: WIP
    # Get the tags for this Feature Group
    my_tags = feature_store.get_feature_group_tags(my_group)
    print(f"Tags: {my_tags}")

    # Set the tags for this table
    feature_store.set_feature_group_tags(my_group, ['public', 'solubility'])

    # Refresh the connector to get the latest info from AWS Feature Store
    feature_store.refresh()

    # Get the tags for this table
    my_tags = feature_store.get_feature_group_tags(my_group)
    print(f"Tags: {my_tags}")

    # Set the tags for this table
    feature_store.add_feature_group_tags(my_group, ['aqsol', 'smiles'])

    # Refresh the connector to get the latest info from AWS Feature Store
    feature_store.refresh()

    # Get the tags for this table
    tags = feature_store.get_feature_group_tags(my_group)
    print(f"Tags: {tags}")
    """



"""Endpoints: Helper Class for AWS SageMaker Endpoints"""
import sys
import argparse
from sagemaker.session import Session

# SageWorks Imports
from sageworks.aws_service_broker.aws_service_connectors.connector import Connector


class Endpoints(Connector):
    def __init__(self):
        """Endpoints: Helper Class for AWS SageMaker Endpoints"""
        # Call SuperClass Initialization
        super().__init__()

        # Set up our SageMaker Session and SageMaker Client
        self.sm_session = Session()
        self.sm_client = self.sm_session.boto_session.client("sagemaker")

        # Set up our internal data storage
        self.endpoint_data = {}

    def check(self) -> bool:
        """Check if we can reach/connect to this AWS Service"""
        try:
            self.sm_client.list_model_package_groups()
            return True
        except Exception as e:
            self.log.critical(f"Error connecting to AWS Model Registry Service: {e}")
            return False

    def refresh(self):
        """Load/reload the tables in the database"""
        # Grab all the Endpoint Data from SageMaker
        print("Reading Endpoints from SageMaker...")
        _endpoints = self.sm_client.list_endpoints()['Endpoints']
        _end_names = [_endpoint['EndpointName'] for _endpoint in _endpoints]

        # Get the details for each Model Group and convert to a data structure with direct lookup
        self.endpoint_data = {name: self._retrieve_details(name) for name in _end_names}

    def metadata(self) -> dict:
        """Get all the table information in this database"""
        return self.endpoint_data

    def endpoint_names(self) -> list:
        """Get all the feature group names in this database"""
        return list(self.endpoint_data.keys())

    def endpoint_details(self, endpoint_name: str) -> dict:
        """Get the details for a specific feature group"""
        return self.endpoint_data.get(endpoint_name)

    def s3_storage(self, endpoint_name: str) -> str:
        """Get the S3 Location for a specific feature group"""
        return 'TBD Later'

    def _retrieve_details(self, endpoint_name: str) -> dict:
        """Internal: Do not call this method directly, use endpoint_details() instead"""

        # Grab the Model Group details from the AWS Model Registry
        details = self.sm_client.describe_endpoint(EndpointName=endpoint_name)
        return details


if __name__ == '__main__':
    from pprint import pprint

    # Collect args from the command line
    parser = argparse.ArgumentParser()
    args, commands = parser.parse_known_args()

    # Check for unknown args
    if commands:
        print('Unrecognized args: %s' % commands)
        sys.exit(1)

    # Create the class and get the AWS Model Registry details
    my_endpoints = Endpoints()
    my_endpoints.refresh()

    # List the Endpoint Names
    print('Endpoints:')
    for end_name in my_endpoints.endpoint_names():
        print(f"\t{end_name}")

    # Get the details for a specific Endpoint
    endpoint_info = my_endpoints.endpoint_details(end_name)
    pprint(endpoint_info)



"""A confusion matrix component"""
from dash import dcc
import plotly.figure_factory as ff


def create_figure(matrix: list[list[float]]):
    """Create the Figure for the Confusion Matrix"""

    x = ['low', 'medium', 'high']
    y = ['high', 'medium', 'low']
    z = matrix

    # change each element of z to type string for annotations
    z_text = [[str(y) for y in x] for x in z]

    # set up figure
    fig = ff.create_annotated_heatmap(z, x=x, y=y, annotation_text=z_text, zmin=0, zmax=1)  # colorscale='Viridis')

    # add title
    fig.update_layout(title_text='<i><b>Confusion matrix</b></i>',
                      # xaxis = dict(title='x'),
                      # yaxis = dict(title='x')
                      )

    # add custom x-axis title
    fig.add_annotation(dict(font=dict(size=14),
                            x=0.5,
                            y=-0.15,
                            showarrow=False,
                            text="Predicted value",
                            xref="paper",
                            yref="paper"))

    # add custom y-axis title
    fig.add_annotation(dict(font=dict(size=14),
                            x=-0.25,
                            y=0.5,
                            showarrow=False,
                            text="Actual value",
                            textangle=-90,
                            xref="paper",
                            yref="paper"))

    # adjust margins to make room for yaxis title
    fig.update_layout(margin=dict(t=50, l=80))

    # add colorbar
    fig['data'][0]['showscale'] = True
    return fig


def create(matrix: list[list[float]]) -> dcc.Graph:
    """Create a Confusion Matrix"""

    # Generate a figure and wrap it in a Dash Graph Component
    return dcc.Graph(id='confusion_matrix', figure=create_figure(matrix))



"""A table component"""
from dash import dash_table
import pandas as pd


def create(table_id: str, df: pd.DataFrame, show_columns: list[str] = None, row_select='single') -> dash_table:
    """Create a Table"""

    # To select rows we need to set up an ID for each row
    df['id'] = df.index

    # Only show these columns
    if not show_columns:
        show_columns = df.columns.to_list()
        show_columns.remove('id')

    table = dash_table.DataTable(
        id=table_id,
        data=df.to_dict('records'),
        columns=[{"name": i, "id": i} for i in df.columns if i in show_columns],
        # style_as_list_view=True,
        sort_action='native',
        # row_selectable=row_select,
        cell_selectable=False,
        # selected_rows=[0],
        # fixed_rows={'headers': True},
        style_table={'maxHeight': '300px', 'overflowY': 'auto'}
    )
    return table


# Storage
"""
style_header={
    'backgroundColor': 'rgb(120, 170, 120)',
    'color': 'rgb(20, 20, 20)'
},
style_data={
    'backgroundColor': 'rgb(30, 30, 30)',
    'color': 'rgb(180, 180, 220)'
}
"""



"""S3LightToDataSource: Class to move LIGHT S3 Files into a SageWorks DataSource"""
import awswrangler as wr
import json

# Local imports
from sageworks.transforms.transform import Transform, TransformInput, TransformOutput
from sageworks.artifacts.data_sources.athena_source import AthenaSource


class S3LightToDataSource(Transform):
    def __init__(self, input_uuid=None, output_uuid=None):
        """S3LightToDataSource: Class to move LIGHT S3 Files into a SageWorks DataSource"""

        # Call superclass init
        super().__init__(input_uuid, output_uuid)

        # Set up all my instance attributes
        self.input_type = TransformInput.S3_OBJECT
        self.output_type = TransformOutput.DATA_SOURCE

    def input_size_mb(self) -> int:
        """ Get the size of the input S3 object in MBytes"""
        size_in_bytes = wr.s3.size_objects(self.input_uuid)[self.input_uuid]
        size_in_mb = round(size_in_bytes/1_000_000)
        return size_in_mb

    def transform(self, overwrite: bool = True):
        """Convert the CSV data into Parquet Format in the SageWorks S3 Bucket, and
           store the information about the data to the AWS Data Catalog sageworks database"""

        # Sanity Check for S3 Object size
        object_megabytes = self.input_size_mb()
        if object_megabytes > 100:
            self.log.error(f"S3 Object too big ({object_megabytes} MBytes): Use the S3HeavyToDataSource class!")
            return

        # Add some tags here
        tags = ['sageworks', 'public']

        # Read in the S3 CSV as a Pandas DataFrame
        df = wr.s3.read_csv(self.input_uuid, low_memory=False)

        # Create the Output Parquet file S3 Storage Path
        s3_storage_path = f"{self.data_source_s3_path}/{self.output_uuid}"

        # Write out the DataFrame to Parquet/DataStore/Athena
        wr.s3.to_parquet(df, path=s3_storage_path, dataset=True, mode='overwrite',
                         database=self.data_catalog_db, table=self.output_uuid,
                         description=f'SageWorks data source: {self.output_uuid}',
                         filename_prefix=f'{self.output_uuid}_',
                         parameters={'tags': json.dumps(tags)},
                         partition_cols=None)  # FIXME: Have some logic around partition columns


# Simple test of the S3LightToDataSource functionality
def test():
    """Test the S3LightToDataSource Class"""
    import pandas as pd

    # Create my Data Loader
    output_uuid = 'aqsol_data'
    my_loader = S3LightToDataSource('s3://sageworks-incoming-data/aqsol_public_data.csv', output_uuid)

    # Store this data as a SageWorks DataSource
    my_loader.transform()

    # Grab the output and print out it's AWS UUID and tags
    output = AthenaSource(output_uuid)
    print(f"UUID: {output.uuid()}   TAGS: {output.tags()}")

    # Now query the data source and print out the resulting dataframe
    df = output.query(f"select * from {output_uuid} limit 5")

    # Setup Pandas output options
    pd.set_option('display.max_colwidth', 15)
    pd.set_option('display.max_columns', 15)
    pd.set_option('display.width', 1000)

    # Show the dataframe
    print(df)


if __name__ == "__main__":
    test()



"""Layout for the model scoreboard"""
from dash import Dash, html
import dash_bootstrap_components as dbc


def artifact_layout(app: Dash, components: dict) -> html.Div:
    layout = html.Div(children=[
        dbc.Row(html.H2(app.title)),
        # Just put all the tables in as Rows for Now (do something fancy later)
        dbc.Row(html.H3('Incoming Data'), style={'padding': '50px 0px 0px 0px'}),
        dbc.Row(components['incoming_data']),
        dbc.Row(html.H3('Data Sources'), style={'padding': '50px 0px 0px 0px'}),
        dbc.Row(components['data_sources']),
        dbc.Row(html.H3('Feature Sets'), style={'padding': '50px 0px 0px 0px'}),
        dbc.Row(components['feature_sets']),
        dbc.Row(html.H3('Models'), style={'padding': '50px 0px 0px 0px'}),
        dbc.Row(components['models']),
        dbc.Row(html.H3('Endpoints'), style={'padding': '50px 0px 0px 0px'}),
        dbc.Row(components['endpoints']),
        ], style={'margin': '30px'})
    return layout



"""Artifact: Abstract Base Class for all Artifact classes in SageWorks.
                Artifacts simply reflect and aggregate one or more AWS Services"""
from abc import ABC, abstractmethod
from datetime import datetime
import logging

from sageworks.utils.logging import logging_setup

# Setup Logging
logging_setup()


class Artifact(ABC):
    def __init__(self):
        """Artifact: Abstract Base Class for all Artifact classes in SageWorks.
                        Artifacts simply reflect and aggregate one or more AWS Services"""
        self.log = logging.getLogger(__name__)

        # FIXME: We should have this come from AWS or Config
        self.data_catalog_db = 'sageworks'
        self.data_source_s3_path = 's3://sageworks-data-sources'
        self.feature_sets_s3_path = 's3://sageworks-feature-sets'

    @abstractmethod
    def check(self) -> bool:
        """Does the Artifact exist? Can we connect to it?"""
        pass

    @abstractmethod
    def uuid(self) -> int:
        """Return the unique SageWorks identifier for this artifact"""
        pass

    @abstractmethod
    def size(self) -> int:
        """Return the size of this artifact in MegaBytes"""
        pass

    @abstractmethod
    def created(self) -> datetime:
        """Return the datetime when this artifact was created"""
        pass

    @abstractmethod
    def modified(self) -> datetime:
        """Return the datetime when this artifact was last modified"""
        pass

    @abstractmethod
    def meta(self):
        """Get the metadata for this artifact"""
        pass

    @abstractmethod
    def tags(self):
        """Get the tags for this artifact"""
        pass

    @abstractmethod
    def add_tag(self, tag):
        """Add a tag to this artifact"""
        pass

    @abstractmethod
    def aws_url(self):
        """AWS console/web interface for this artifact"""
        pass



"""ArtifactsSummary pulls All the metadata from the AWS Service Broker and organizes/summarizes it"""
import sys
import argparse

import pandas as pd

# SageWorks Imports
from sageworks.views.view import View
from sageworks.aws_service_broker.aws_service_broker import ServiceCategory


class ArtifactsSummary(View):
    def __init__(self):
        """ArtifactsSummary pulls All the metadata from the AWS Service Broker and organizes/summarizes it"""
        # Call SuperClass Initialization
        super().__init__()

        # Get AWS Service information for ALL the categories (data_source, feature_set, endpoints, etc)
        self.service_info = self.aws_broker.get_all_metadata()

        # Summary data for ALL the AWS Services
        self.summary_data = {}

    def check(self) -> bool:
        """Can we connect to this view/service?"""
        True  # I'm great, thx for asking

    def refresh(self) -> bool:
        """Refresh data/metadata associated with this view"""
        self.service_info = self.aws_broker.get_all_metadata()

    def view_data(self) -> dict:
        """Get all the data that's useful for this view

            Returns:
                dict: Dictionary of Pandas Dataframes, i.e. {'INCOMING_DATA', pd.DataFrame}
           """

        # We're filling in Summary Data for all the AWS Services
        self.summary_data['INCOMING_DATA'] = self.incoming_data_summary()
        self.summary_data['DATA_SOURCES'] = self.data_sources_summary()
        self.summary_data['FEATURE_SETS'] = self.feature_sets_summary()
        self.summary_data['MODELS'] = self.models_summary()
        self.summary_data['ENDPOINTS'] = self.endpoints_summary()
        return self.summary_data

    def incoming_data_summary(self):
        """Get summary data about data in the incoming-data S3 Bucket"""
        data = self.service_info[ServiceCategory.INCOMING_DATA]
        data_summary = []
        for name, info in data.items():
            summary = {'Name': name,
                       'Size': str(info.get('ContentLength', '-')),
                       'LastModified': self.datetime_string(info.get('LastModified')),
                       'ContentType': str(info.get('ContentType', '-')),
                       'ServerSideEncryption': info.get('ServerSideEncryption', '-'),
                       'Tags': str(info.get('tags', '-'), )}
            data_summary.append(summary)

        return pd.DataFrame(data_summary)

    def data_sources_summary(self):
        """Get summary data about the SageWorks DataSources"""
        data = self.service_info[ServiceCategory.DATA_CATALOG]
        data_summary = []
        for database, db_info in data.items():
            for name, info in db_info.items():
                summary = {'Name': name,
                           'Catalog DB': info.get('DatabaseName', '-'),
                           'Size': str(info.get('ContentLength', '-')),
                           'Created': self.datetime_string(info.get('CreateTime')),
                           'LastModified': self.datetime_string(info.get('UpdateTime')),
                           'Num Columns': self.num_columns(info),
                           'DataLake': info.get('IsRegisteredWithLakeFormation', '-'),
                           'Tags': str(info.get('tags', '-'), )}
                data_summary.append(summary)

        return pd.DataFrame(data_summary)

    def feature_sets_summary(self):
        """Get summary data about the SageWorks FeatureSets"""
        data = self.service_info[ServiceCategory.FEATURE_STORE]
        data_summary = []
        for feature_group, group_info in data.items():
            summary = {'Feature Group': group_info['FeatureGroupName'],
                       'Catalog DB': group_info['OfflineStoreConfig']['DataCatalogConfig'].get('Database', '-'),
                       'Feature ID': group_info['RecordIdentifierFeatureName'],
                       'Feature EventTime': group_info['EventTimeFeatureName'],
                       'Online': str(group_info.get('OnlineStoreConfig', {}).get('EnableOnlineStore', 'False')),
                       'Created': self.datetime_string(group_info.get('CreationTime')),
                       'LastModified': self.datetime_string(group_info.get('LastModified')),
                       'Tags': str(group_info.get('tags', '-'), )}
            data_summary.append(summary)

        return pd.DataFrame(data_summary)

    def models_summary(self):
        """Get summary data about the SageWorks Models"""
        data = self.service_info[ServiceCategory.MODELS]
        data_summary = []
        for model_group, model_list in data.items():
            # Special Case for Model Groups without any Models
            if not model_list:
                summary = {'Model Group': model_group}
                data_summary.append(summary)
                continue

            # Get Summary information for each model in the model_list
            for model in model_list:
                summary = {'Model Group': model['ModelPackageGroupName'],
                           'Version': model['ModelPackageVersion'],
                           'Description': model['ModelPackageDescription'],
                           'Created': self.datetime_string(model.get('CreationTime')),
                           'LastModified': self.datetime_string(model.get('LastModifiedTime')),
                           'Status': str(model.get('ModelApprovalStatus', ' - ')),
                           'Tags': str(model.get('tags', '-'), )}
                data_summary.append(summary)

        return pd.DataFrame(data_summary)

    def endpoints_summary(self):
        """Get summary data about the SageWorks Endpoints"""
        data = self.service_info[ServiceCategory.ENDPOINTS]
        data_summary = []

        # Get Summary information for each endpoint
        for endpoint, endpoint_info in data.items():
            summary = {'Name': endpoint_info['EndpointName'],
                       'Status': endpoint_info['EndpointStatus'],
                       'Description': 'TBD',
                       'Created': self.datetime_string(endpoint_info.get('CreationTime')),
                       'LastModified': self.datetime_string(endpoint_info.get('LastModifiedTime')),
                       'DataCapture': str(endpoint_info.get('DataCaptureConfig', {}).get('EnableCapture', 'False')),
                       'SamplingPercent': str(endpoint_info.get('DataCaptureConfig', {}).get('CurrentSamplingPercentage', '-')),
                       'Tags': str(endpoint_info.get('tags', '-'), )}
            data_summary.append(summary)

        return pd.DataFrame(data_summary)


    @staticmethod
    def num_columns(data_info):
        """Helper: Compute the number of columns from the storage descriptor data"""
        try:
            return len(data_info['StorageDescriptor']['Columns'])
        except KeyError:
            return '-'

    @staticmethod
    def datetime_string(datetime_obj):
        """Helper: Convert DateTime Object into a nice string"""
        if datetime_obj is None:
            return '-'
        # Date + Hour Minute
        return datetime_obj.strftime("%Y-%m-%d %H:%M")


if __name__ == '__main__':

    # Collect args from the command line
    parser = argparse.ArgumentParser()
    args, commands = parser.parse_known_args()

    # Check for unknown args
    if commands:
        print('Unrecognized args: %s' % commands)
        sys.exit(1)

    # Create the class and get the AWS Model Registry details
    artifact_view = ArtifactsSummary()

    # List the Endpoint Names
    print('ArtifactsSummary:')
    for category, df in artifact_view.view_data().items():
        print(f"\n{category}")
        print(df.head())



"""A Component for model details/information"""
from dash import dcc


def create_markdown(model_info: dict):
    """Create the Markdown for the Model Details"""

    markdown_template = """
        #### {{model_name}}
        - **Created:** {{date_created}}
        - **Training Data:**
            - Nightly ({{date_created}})
            - {{training_data}}
        - **Features Set:** {{feature_set}}
        - **Type:** {{model_type}}
        - **Algorithm:** XGBoost
        - **Endpoints:**
            - {{model_name}}-80
            - {{model_name}}-100

        ##### Model Scores and Metrics

        All model scores and metrics are based on 80/20 train/test splits of the **Training Data** (see above).
        The project will be improving upon these existing reference models as the project progresses.
    """
    # Hack Logic
    if 'regress' in model_info['model_name']:
        model_type = 'Regression/Numerical'
    else:
        model_type = 'Classification/Categorical'

    # Replace all the template fields
    markdown = markdown_template.replace('{{model_name}}', model_info['model_name'])
    markdown = markdown.replace('{{date_created}}', str(model_info['date_created'])[:10])
    markdown = markdown.replace('{{training_data}}', model_info['training_data'])
    markdown = markdown.replace('{{feature_set}}', model_info['feature_set'])
    markdown = markdown.replace('{{model_type}}', model_type)

    return markdown


def create(model_info: dict) -> dcc.Markdown:
    """Create a Markdown Component for the Model Details"""

    # Generate a figure and wrap it in a Dash Graph Component
    return dcc.Markdown(id='model_details', children=create_markdown(model_info))



"""FeaturesToPandas: Class to transform a FeatureSet into a Pandas DataFrame"""
import logging
import pandas as pd

# Local imports
from sageworks.utils.logging import logging_setup
from sageworks.transforms.transform import Transform, TransformInput, TransformOutput
from sageworks.artifacts.data_sources.athena_source import AthenaSource

# Setup Logging
logging_setup()


class FeaturesToPandas(Transform):
    def __init__(self):
        """FeaturesToPandas: Class to transform a FeatureSet into a Pandas DataFrame"""

        # Call superclass init
        super().__init__()

        # Set up all my instance attributes
        self.log = logging.getLogger(__name__)
        self.input_uuid = None
        self.input_data_source = None
        self.output_df = None
        self.data_catalog_db = 'sageworks'

    def input_type(self) -> TransformInput:
        """What Input Type does this Transform Consume"""
        return TransformInput.FEATURE_SET

    def output_type(self) -> TransformOutput:
        """What Output Type does this Transform Produce"""
        return TransformOutput.PANDAS_DF

    def set_input_uuid(self, input_uuid: str):
        self.input_uuid = input_uuid

    def set_output_uuid(self, uuid: str):
        """Not Implemented: Just satisfying the Transform abstract method requirements"""
        pass

    def get_output(self) -> pd.DataFrame:
        """Get the DataFrame Output from this Transform"""
        return self.output_df

    def validate_input(self) -> bool:
        """Validate the Input for this Transform"""
        self.input_data_source = AthenaSource(self.input_uuid, self.data_catalog_db)
        return self.input_data_source.check()

    def transform(self, max_rows=100000):
        """Convert the DataSource into a Pandas DataFrame"""

        # Validate our input and get the number of rows in the DataSource
        self.validate_input()
        num_rows = self.input_data_source.num_rows()

        # If the data source has more rows than max_rows, do a sample query
        if num_rows > max_rows:
            percentage = round(max_rows*100.0/num_rows)
            self.log.warning(f"DataSource has {num_rows} rows.. sampling down to {max_rows}...")
            query = f"SELECT * FROM {self.input_uuid} TABLESAMPLE BERNOULLI({percentage})"
        else:
            query = f"SELECT * FROM {self.input_uuid}"
        self.output_df = self.input_data_source.query(query)


# Simple test of the FeaturesToPandas functionality
def test():
    """Test the FeaturesToPandas Class"""

    # Setup Pandas output options
    pd.set_option('display.max_colwidth', 15)
    pd.set_option('display.max_columns', 15)
    pd.set_option('display.width', 1000)

    # Grab a Data Source
    data_uuid = 'aqsol_data'

    # Create the DataSource to DF Transform
    data_to_df = FeaturesToPandas()
    data_to_df.set_input_uuid(data_uuid)

    # Transform the DataSource into a Pandas DataFrame (with max_rows = 1000)
    data_to_df.transform(max_rows=1000)

    # Grab the output and show it
    my_df = data_to_df.get_output()
    print(my_df)


if __name__ == "__main__":
    test()



"""Callbacks/Connections in the Web User Interface"""
import dash
from dash import Dash
from dash.dependencies import Input, Output

# SageWorks Imports
from web_interfaces.web_tools.components.model_data import ModelData
from web_interfaces.web_tools.components import feature_importance, confusion_matrix, model_details, feature_details


# Highlights the selected row in the table
def table_row_select(app: Dash, table_name: str):
    @app.callback(
        Output(table_name, "style_data_conditional"),
        Input(table_name, "derived_viewport_selected_row_ids"),
    )
    def style_selected_rows(selected_rows):
        if selected_rows is None:
            return dash.no_update
        foo = [
            {"if": {"filter_query": "{{id}} ={}".format(i)}, "backgroundColor": "rgb(200, 220, 200)"}
            for i in selected_rows
        ]
        return foo


# Updates the feature importance and confusion matrix figures when a model is selected
def update_figures(app: Dash, model_data: ModelData):
    @app.callback(
        [Output('feature_importance', "figure"), Output('confusion_matrix', "figure")],
        Input('model_table', "derived_viewport_selected_row_ids"),
    )
    def generate_new_figures(selected_rows):
        print(f'Selected Rows: {selected_rows}')

        # If there's no selection we're going to return figures for the first row (0)
        if not selected_rows:
            selected_rows = [0]

        # Grab the data for this row
        model_row_index = selected_rows[0]

        # Generate a figure for the feature importance component
        feature_info = model_data.get_model_feature_importance(model_row_index)
        feature_figure = feature_importance.create_figure(feature_info)

        # Generate a figure for the confusion matrix component
        c_matrix = model_data.get_model_confusion_matrix(model_row_index)
        matrix_figure = confusion_matrix.create_figure(c_matrix)

        # Now return both of the new figures
        return [feature_figure, matrix_figure]


# Updates the model details when a model row is selected
def update_model_details(app: Dash, model_data: ModelData):
    @app.callback(
        Output('model_details', "children"),
        Input('model_table', "derived_viewport_selected_row_ids"),
    )
    def generate_new_markdown(selected_rows):
        print(f'Selected Rows: {selected_rows}')

        # If there's no selection we're going to return the model details for the first row (0)
        if not selected_rows:
            selected_rows = [0]

        # Grab the data for this row
        model_row_index = selected_rows[0]

        # Generate new Details (Markdown) for the selected model
        model_info = model_data.get_model_details(model_row_index)
        model_markdown = model_details.create_markdown(model_info)

        # Return the details/markdown for this model
        return model_markdown


# Updates the feature details when a model row is selected
def update_feature_details(app: Dash, model_data: ModelData):
    @app.callback(
        Output('feature_details', "children"),
        Input('model_table', "derived_viewport_selected_row_ids"),
    )
    def generate_new_markdown(selected_rows):
        print(f'Selected Rows: {selected_rows}')

        # If there's no selection we're going to return the feature details for the first row (0)
        if not selected_rows:
            selected_rows = [0]

        # Grab the data for this row
        model_row_index = selected_rows[0]

        # Generate new Details (Markdown) for the features for this model
        feature_info = model_data.get_model_feature_importance(model_row_index)
        feature_markdown = feature_details.create_markdown(feature_info)

        # Return the details/markdown for these features
        return feature_markdown



"""View: View in the database sense: Pulls from the AWS Service Broker and does slice and dice"""
from abc import ABC, abstractmethod

import logging

# SageWorks Imports
from sageworks.aws_service_broker.aws_service_broker import AWSServiceBroker
from sageworks.utils.logging import logging_setup

# Set up logging
logging_setup()


class View(ABC):
    def __init__(self, database_scope='all'):
        """View: View in the database sense: Pulls from the AWS Service Broker and does slice and dice"""
        self.log = logging.getLogger(__name__)

        # Grab an AWS Metadata Broker object for pulling AWS Service information
        self.aws_broker = AWSServiceBroker(database_scope=database_scope)

    @abstractmethod
    def check(self) -> bool:
        """Can we connect to this view/service?"""
        pass

    @abstractmethod
    def refresh(self) -> bool:
        """Refresh data/metadata associated with this view"""
        pass

    @abstractmethod
    def view_data(self) -> dict:
        """Return all the data that's useful for this view"""
        pass



"""HelloWorld: A SageWorks HelloWorld Application"""
from dash import Dash
import dash_bootstrap_components as dbc


# SageWorks Imports
from sageworks.views.artifacts_summary import ArtifactsSummary
from sageworks.web_components import table

# Local Imports
import layout


# Note: The 'app' and 'server' objects need to be at the top level since NGINX/uWSGI needs to
#       import this file and use the server object as an ^entry-point^ into the Dash Application Code

# Create our Dash Application
app = Dash(title='SageWork Artifacts', external_stylesheets=[dbc.themes.BOOTSTRAP])
# app = Dash(title='Hello World Application', external_stylesheets=[dbc.themes.DARKLY])
server = app.server


def setup_artifact_viewer():

    # Set Default Template for figures
    # load_figure_template('darkly')

    # Grab a view that gives us a summary of all the artifacts currently in SageWorks
    sageworks_artifacts = ArtifactsSummary()
    artifacts_summary = sageworks_artifacts.view_data()

    # Just a bunch of tables for now :)
    tables = {}
    for service_category, artifact_info_df in artifacts_summary.items():

        # Grab the Artifact Information DataFrame for each AWS Service
        tables[service_category] = table.create(service_category, artifact_info_df)

    # Create our components
    components = {
        'incoming_data': tables['INCOMING_DATA'],
        'data_sources': tables['DATA_SOURCES'],
        'feature_sets': tables['FEATURE_SETS'],
        'models': tables['MODELS'],
        'endpoints': tables['ENDPOINTS']
    }

    # Setup up our application layout
    app.layout = layout.artifact_layout(app, components)

    # Setup our callbacks/connections
    """
    callbacks.table_row_select(app, 'model_table')
    callbacks.update_figures(app, df)
    callbacks.update_model_details(app, df)
    callbacks.update_feature_details(app, df)
    """


# Now actually set up the scoreboard
setup_artifact_viewer()


if __name__ == '__main__':
    # Run our web application in TEST mode
    # Note: This 'main' is purely for running/testing locally
    app.run_server(host='0.0.0.0', port=8080, debug=True)



"""A Component for feature details/information"""
from dash import dcc


def create_markdown(feature_info: dict):
    """Create the Markdown for the Feature Details"""

    # Sort the features
    feature_list = sorted(feature_info.items(), key=lambda x: x[1], reverse=True)

    markdown = ""
    for (feature, value) in feature_list:
        markdown += f"- **{feature}** ({value}): Cool stuff about descriptor {feature} that we get later [{feature}](https://www.rdkit.org/docs/source/rdkit.Chem.Crippen.html#rdkit.Chem.Crippen.MolLogP)\n"

    # Add the additional info section
    markdown += """
    ##### Additional Information on Features

    For additional information on the features used in this model please
    see
    - [RDKIT Descriptors](https://www.rdkit.org/docs/source/rdkit.Chem.Descriptors.html)
    - [Mordred Descriptors](https://mordred-descriptor.github.io/documentation/master/descriptors.html)
    """
    return markdown


def create(feature_info: dict) -> dcc.Markdown:
    """Create a Markdown Component for the Feature Details"""

    # Generate a figure and wrap it in a Dash Graph Component
    return dcc.Markdown(id='feature_details', children=create_markdown(feature_info))



"""ArtifactsSummary pulls All the metadata from the AWS Service Broker and organizes/summarizes it"""
import sys
import argparse

import pandas as pd

# SageWorks Imports
from sageworks.views.view import View
from sageworks.aws_service_broker.aws_service_broker import ServiceCategory
from sageworks.artifacts.data_sources.athena_source import AthenaSource


class ArtifactsSummary(View):
    def __init__(self):
        """ArtifactsSummary pulls All the metadata from the AWS Service Broker and organizes/summarizes it"""
        # Call SuperClass Initialization
        super().__init__()

        # Get AWS Service information for ALL the categories (data_source, feature_set, endpoints, etc)
        self.service_info = self.aws_broker.get_all_metadata()

        # Summary data for ALL the AWS Services
        self.summary_data = {}

    def check(self) -> bool:
        """Can we connect to this view/service?"""
        True  # I'm great, thx for asking

    def refresh(self) -> bool:
        """Refresh data/metadata associated with this view"""
        self.service_info = self.aws_broker.get_all_metadata()

    def view_data(self) -> dict:
        """Get all the data that's useful for this view

            Returns:
                dict: Dictionary of Pandas Dataframes, i.e. {'INCOMING_DATA', pd.DataFrame}
           """

        # We're filling in Summary Data for all the AWS Services
        self.summary_data[ServiceCategory.INCOMING_DATA] = self.incoming_data_summary()
        self.summary_data[ServiceCategory.DATA_CATALOG] = self.data_sources_summary()
        self.summary_data[ServiceCategory.FEATURE_STORE] = self.incoming_data_summary()
        self.summary_data[ServiceCategory.MODELS] = self.incoming_data_summary()
        self.summary_data[ServiceCategory.ENDPOINTS] = self.incoming_data_summary()
        return self.summary_data

    def incoming_data_summary(self):
        """Get summary data about data in the incoming-data S3 Bucket"""
        data = self.service_info[ServiceCategory.INCOMING_DATA]
        data_summary = []
        for name, info in data.items():
            summary = {'Name': name,
                       'Size': str(info.get('ContentLength', '-')),
                       'LastModified': str(info.get('LastModified', '-')),
                       'ContentType': str(info.get('ContentType', '-')),
                       'ServerSideEncryption': info.get('ServerSideEncryption', '-'),
                       'Tags': str(info.get('tags', '-'), )}
            data_summary.append(summary)

        return pd.DataFrame(data_summary)

    def data_sources_summary(self):
        """Get summary data about the SageWorks DataSources"""
        data = self.service_info[ServiceCategory.DATA_CATALOG]
        data_summary = []
        for database, db_info in data.items():
            for name, info in db_info.items():

                # We pull from the Concrete Class to get nice API for metadata
                athena_source = AthenaSource(name, database=info.get('DatabaseName', 'sageworks'))
                summary = {'Name': athena_source.uuid(),
                           'Database': athena_source.data_catalog_db,
                           'Size': '-',  # athena_source.size(),  # This takes TOO LONG
                           'NumColumns': athena_source.num_columns(),
                           'NumRows': '-',  # athena_source.num_rows(),  # This takes TOO LONG
                           'Created': athena_source.created(),
                           'LastModified': athena_source.modified(),
                           'Location': athena_source.s3_storage_location(),
                           'AmazonURL': athena_source.aws_url(),
                           'DataLake Registered': info.get('IsRegisteredWithLakeFormation', '-'),
                           'Tags': str(athena_source.tags())}
                data_summary.append(summary)

        return pd.DataFrame(data_summary)

    def feature_store(self):
        """Get summary data about the SageWorks DataSources"""

        # FIXME: Fill in with real logic
        return self.data_sources_summary()

    def models(self):
        """Get summary data about the SageWorks DataSources"""

        # FIXME: Fill in with real logic
        return self.data_sources_summary()

    def endpoints(self):
        """Get summary data about the SageWorks DataSources"""

        # FIXME: Fill in with real logic
        return self.data_sources_summary()


if __name__ == '__main__':

    # Collect args from the command line
    parser = argparse.ArgumentParser()
    args, commands = parser.parse_known_args()

    # Check for unknown args
    if commands:
        print('Unrecognized args: %s' % commands)
        sys.exit(1)

    # Create the class and get the AWS Model Registry details
    artifact_view = ArtifactsSummary()

    # List the Endpoint Names
    print('ArtifactsSummary:')
    for category, df in artifact_view.view_data().items():
        print(f"\n{category}")
        print(df.head())



"""FeatureSpider: A Spider for data/feature investigation and QA"""
# License: Apache 2.0 ©SuperCowPowers LLC
import numpy as np
import pandas as pd
from sklearn.neighbors import KNeighborsRegressor
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import StandardScaler


# Feature Spider Class
class FeatureSpider:
    def __init__(self, df: pd.DataFrame, features: list, id_column: str, target: str):

        # Check for expected columns (used later)
        for column in [id_column, target] + features:
            if column not in df.columns:
                print(f'DataFrame does not have required {column} Column!')
                return

        # Set internal vars that are used later
        self.df = df
        self.id_column = id_column
        self.target = target
        self.features = features

        # Build our KNN model pipeline with StandardScalar
        knn = KNeighborsRegressor(n_neighbors=10, weights='distance')
        self.pipe = make_pipeline(StandardScaler(), knn)

        # Fit Model on features and target
        y = df[target]
        X = df[features]
        self.pipe.fit(X, y)

        # Grab the Standard Scalar and KNN from the pipeline model
        # Note: These handles need to be constructed after the fit
        self.scalar = self.pipe['standardscaler']
        self.knn = self.pipe['kneighborsregressor']

    def get_feature_matrix(self):
        """Return the KNN Model Internal Feature Matrix"""
        return self.knn._fit_X

    def predict(self, pred_df: pd.DataFrame) -> list:
        """Provide a prediction from the KNN Pipeline model (knn_prediction)"""
        return self.pipe.predict(pred_df[self.features])

    def confidence_scores(self, pred_df: pd.DataFrame, model_preds=None) -> list:
        """Compute Confidence Scores for each Prediction"""

        # Get all the KNN information relevant to this calculation
        neighbor_info = self.neighbor_info(pred_df)

        # Handles for all the relevant info
        knn_preds = neighbor_info['knn_prediction']
        target_values = neighbor_info['knn_target_values']
        distances = neighbor_info['knn_distances']

        # We can score confidence even if we don't have model predictions (less good)
        if model_preds is None:
            model_preds = knn_preds
            stddev_multiplier = 1.5
        else:
            stddev_multiplier = 1.0

        # Now a big loop over all these values to compute the confidence scores
        confidence_scores = []
        for pred, knn_pred, str_val_list, str_dist_list in zip(model_preds, knn_preds, target_values, distances):
            # Each of these is a string of a list (yes a bit cheesy)
            vals = [float(val) for val in str_val_list.split(', ')]
            _ = [float(dis) for dis in str_dist_list.split(', ')]  # dist current not used

            # Compute stddev of the target values
            knn_stddev = np.std(vals)

            # Confidence Score
            conf = 0.5 * (2.0 - abs(float(pred) - float(knn_pred)))
            conf -= knn_stddev * stddev_multiplier

            # Confidence score has min-max of 0-1
            conf = min(max(conf, 0), 1)

            confidence_scores.append(conf)

        # Return the confidence scores
        return confidence_scores

    def neighbor_info(self, pred_df: pd.DataFrame) -> pd.DataFrame:
        """Provide information on the neighbors (prediction, knn_target_values, knn_distances)"""

        # Make sure we have all the features
        if not set(self.features) <= set(pred_df.columns):
            print(f'DataFrame does not have required features: {self.features}')
            return None

        # Run through scaler
        x_scaled = self.scalar.transform(pred_df[self.features])

        # Add the data to a copy of the dataframe
        results_df = pd.DataFrame()
        results_df['knn_prediction'] = self.knn.predict(x_scaled)

        # Get the Neighbors Information
        neigh_dist, neigh_ind = self.knn.kneighbors(x_scaled)
        target_values = self.knn._y[neigh_ind]

        # Note: We're assuming that the Neighbor Index is the same order/cardinality as the dataframe
        results_df['knn_target_values'] = [', '.join([str(val) for val in values]) for values in target_values]
        results_df['knn_distances'] = [', '.join([str(dis) for dis in distances]) for distances in neigh_dist]

        return results_df

    def neighbor_ids(self, pred_df) -> pd.DataFrame:
        """Provide id, features for the neighbors (knn_ids, knn_features)"""

        # Run through scaler
        x_scaled = self.scalar.transform(pred_df[self.features])

        # Add the data to a copy of the dataframe
        results_df = pred_df.copy()

        # Neighbor ID and feature lookups
        neigh_dist, neigh_ind = self.knn.kneighbors(x_scaled)
        results_df['knn_ids'] = [', '.join(self.df.iloc[index][self.id_column] for index in indexes) for indexes in neigh_ind]
        results_df['knn_features'] = [', '.join(self.df.iloc[index][self.id_column] for index in indexes) for indexes in neigh_ind]
        return results_df

    def coincident(self, target_diff, verbose=True):
        """Convenience method that calls high_gradients with a distance of 0"""
        return self.high_gradients(0.0, target_diff, verbose)

    def high_gradients(self, within_distance: float, target_diff: float, verbose: bool = True) -> list:
        """This basically loops over all the X features in the KNN model
           - Grab the neighbors distances and indices
           - For neighbors `within_distance`* grab target values
           - If target values have a difference > `target_diff`
              - List out the details of the observations and the distance, target diff
           *Note: standardized feature space
        """
        global_htg_set = set()
        for my_index, obs in enumerate(self.knn._fit_X):
            neigh_distances, neigh_indexes = self.knn.kneighbors([obs])
            neigh_distances = neigh_distances[0]  # Returns a list within a list so grab the inner list
            neigh_indexes = neigh_indexes[0]  # Returns a list within a list so grab the inner list
            target_values = self.knn._y[neigh_indexes]

            # Grab the info for this observation
            my_id = self.df.iloc[my_index][self.id_column]
            my_features = self.df.iloc[my_index][self.id_column]
            my_target = self.knn._y[my_index]

            # Loop through the neighbors
            # Note: by definition this observation will be in the neighbors so account for that
            my_htg_set = set()
            for n_index, dist, target in zip(neigh_indexes, neigh_distances, target_values):

                # Skip myself
                if n_index == my_index:
                    continue

                # Compute target differences `within_distance` feature space
                _diff = abs(my_target - target)
                if dist <= within_distance and _diff > target_diff:

                    # Update the individual HTG set for this observation
                    my_htg_set.add((n_index, dist, _diff, target))

                    # Add both (me and my neighbor) to the global high gradient index list
                    global_htg_set.add(my_index)
                    global_htg_set.add(n_index)

            # Okay we've computed our HTG set for this observation
            # Print out all my HTG neighbors if the verbose flag is set
            if verbose and my_htg_set:
                print(f"\nOBSERVATION: {my_id}")
                print(f"\t{my_id}({my_target:.2f}):{my_features}")
                for htg_neighbor, dist, _diff, target in my_htg_set:
                    neighbor_id = self.df.iloc[htg_neighbor][self.id_column]
                    neighbor_features = self.df.iloc[htg_neighbor][self.id_column]
                    print(f"\t{neighbor_id}({target:.2f}):{neighbor_features} TargetD:{_diff:.2f} FeatureD:{dist}")

        # Return the global list of indexes that are part of high target gradient (HTG) pairs
        return list(global_htg_set)


def test():
    """Test for the Feature Spider Class"""
    # Set some pandas options
    pd.set_option('display.max_columns', None)
    pd.set_option('display.width', 1000)

    # Make some fake data
    data = {'ID': ['id_0', 'id_0', 'id_2', 'id_3', 'id_4',
                   'id_5', 'id_6', 'id_7', 'id_8', 'id_9'],
            'feat1': [1.0, 1.0, 1.1, 3.0, 4.0, 1.0, 1.0, 1.1, 3.0, 4.0],
            'feat2': [1.0, 1.0, 1.1, 3.0, 4.0, 1.0, 1.0, 1.1, 3.0, 4.0],
            'feat3': [0.1, 0.1, 0.2, 1.6, 2.5, 0.1, 0.1, 0.2, 1.6, 2.5],
            'price': [31, 60, 62, 40, 20, 31, 61, 60, 40, 20]}
    data_df = pd.DataFrame(data)

    # Create the class and run the taggers
    f_spider = FeatureSpider(data_df, ['feat1', 'feat2', 'feat3'], id_column='ID', target='price')
    preds = f_spider.predict(data_df)
    print(preds)
    coincident = f_spider.coincident(2)
    print('COINCIDENT')
    print(coincident)
    high_gradients = f_spider.high_gradients(2, 2)
    print('\nHIGH GRADIENTS')
    print(high_gradients)

    # Run some neighbor methods
    query_df = data_df[data_df['ID'] == 'id_0'].copy()
    print(f_spider.neighbor_info(query_df))


if __name__ == '__main__':
    test()



import os
from dash import Dash
import dash_bootstrap_components as dbc


# SageWorks Imports
from sageworks.web_components import confusion_matrix, table, scatter_plot
from sageworks.web_components import feature_importance, model_data, model_details, feature_details

# Local Imports
import layout
import callbacks


# Note: The 'app' and 'server' objects need to be at the top level since NGINX/uWSGI needs to
#       import this file and use the server object as an ^entry-point^ into the Dash Application Code

# Create our Dash Application
app = Dash(title='Model Details Viewer', external_stylesheets=[dbc.themes.BOOTSTRAP])
# app = Dash(title='Hello World Application', external_stylesheets=[dbc.themes.DARKLY])
server = app.server


def setup_model_details_view():

    # Set Default Template for figures
    # load_figure_template('darkly')

    # Read in our model data
    data_path = os.path.join(os.path.dirname(__file__), 'data/toy_data.csv')
    model_info = model_data.ModelData(data_path)

    # Create our components
    model_df = model_info.get_model_df()
    model_table = table.create('model_table', model_df, show_columns=['model_name', 'date_created', 'f_scores'])
    details = model_details.create(model_info.get_model_details(0))
    c_matrix = confusion_matrix.create(model_info.get_model_confusion_matrix(0))
    scatter = scatter_plot.create(model_df)
    my_feature_importance = feature_importance.create(model_info.get_model_feature_importance(0))
    my_feature_details = feature_details.create(model_info.get_model_feature_importance(0))
    components = {
        'model_table': model_table,
        'model_details': details,
        'confusion_matrix': c_matrix,
        'scatter_plot': scatter,
        'feature_importance': my_feature_importance,
        'feature_details': my_feature_details
    }

    # Setup up our application layout
    app.layout = layout.scoreboard_layout(app, components)

    # Setup our callbacks/connections
    callbacks.table_row_select(app, 'model_table')
    callbacks.update_figures(app, model_info)
    callbacks.update_model_details(app, model_info)
    callbacks.update_feature_details(app, model_info)


# Now actually set up the model details view
setup_model_details_view()


if __name__ == '__main__':
    # Run our web application in TEST mode
    # Note: This 'main' is purely for running/testing locally
    app.run_server(host='0.0.0.0', port=8080, debug=True)



"""S3Bucket: Class to retrieve object/file information from an AWS S3 Bucket"""
import sys
import argparse
import awswrangler as wr


# SageWorks Imports
from sageworks.aws_service_broker.aws_service_connectors.connector import Connector


# Class to retrieve object/file information from an AWS S3 Bucket
class S3Bucket(Connector):
    """S3Bucket: Class to retrieve object/file information from an AWS S3 Bucket"""
    def __init__(self, bucket: str):
        # Call SuperClass Initialization
        super().__init__()

        # Store our bucket name
        self.bucket = bucket
        self.s3_bucket_data = None

        # Load in the files from the bucket
        self.refresh()

    def check(self) -> bool:
        """Check if we can reach/connect to this AWS Service"""
        try:
            wr.s3.does_object_exist(self.bucket)
            return True
        except Exception as e:
            self.log.critical(f"Could not connect to AWS S3 {self.bucket}: {e}")
            return False

    def refresh(self):
        """Load/reload the files in the bucket"""
        # Grab all the files in this bucket
        self.log.info(f"Reading S3 Bucket: {self.bucket}...")
        _aws_file_info = wr.s3.describe_objects(self.bucket)
        self.s3_bucket_data = {full_path.split('/')[-1]: info for full_path, info in _aws_file_info.items()}

    def metadata(self) -> dict:
        """Get all the metadata for the files in this bucket"""
        return self.s3_bucket_data

    def file_names(self) -> list:
        """Get all the file names in this bucket"""
        return list(self.s3_bucket_data.keys())

    def file_info(self, file: str) -> dict:
        """Get additional info about this specific file"""
        return self.s3_bucket_data[file]


if __name__ == '__main__':
    from pprint import pprint

    # Collect args from the command line
    parser = argparse.ArgumentParser()
    parser.add_argument('--bucket', type=str, default='s3://sageworks-incoming-data', help='AWS S3 Bucket')
    args, commands = parser.parse_known_args()

    # Check for unknown args
    if commands:
        print('Unrecognized args: %s' % commands)
        sys.exit(1)

    # Create the class and get the AWS Data Catalog database info
    s3_bucket = S3Bucket(args.bucket)

    # List files in the bucket
    print(f"{s3_bucket.bucket}")
    for file_name in s3_bucket.file_names():
        print(f"\t{file_name}")

    # Get additional info for a specific file
    my_file_info = s3_bucket.file_info('aqsol_public_data.csv')
    pprint(my_file_info)



"""AWSServiceBroker pulls and collects metadata from a bunch of AWS Services"""
import sys
import argparse
from enum import Enum, auto
import logging

# SageWorks Imports
from sageworks.aws_service_broker.cache import Cache
from sageworks.aws_service_broker.aws_service_connectors.s3_bucket import S3Bucket
from sageworks.aws_service_broker.aws_service_connectors.data_catalog import DataCatalog
from sageworks.aws_service_broker.aws_service_connectors.feature_store import FeatureStore
from sageworks.aws_service_broker.aws_service_connectors.model_registry import ModelRegistry
from sageworks.aws_service_broker.aws_service_connectors.endpoints import Endpoints
from sageworks.utils.logging import logging_setup

# Setup Logging
logging_setup()


# Enumerated types for SageWorks Meta Requests
class ServiceCategory(Enum):
    """Enumerated Types for SageWorks Meta Requests"""
    INCOMING_DATA = auto()
    DATA_CATALOG = auto()
    FEATURE_STORE = auto()
    MODELS = auto()
    ENDPOINTS = auto()


class AWSServiceBroker:

    def __new__(cls, database_scope='sageworks'):
        """AWSServiceBroker Singleton Pattern"""
        if not hasattr(cls, 'instance'):
            print('Creating New AWSServiceBroker Instance...')
            cls.instance = super(AWSServiceBroker, cls).__new__(cls)

            # Class Initialization
            cls.instance.__class_init__(database_scope)

        return cls.instance

    @classmethod
    def __class_init__(cls, database_scope='sageworks'):
        """"AWSServiceBroker pulls and collects metadata from a bunch of AWS Services"""
        cls.log = logging.getLogger(__file__)

        # FIXME: This should be pulled from a config file
        cls.incoming_data_bucket = 's3://sageworks-incoming-data'

        # SageWorks category mapping to AWS Services
        # - incoming_data = S3
        # - data_sources = Data Catalog, Athena
        # - feature_sets = Data Catalog, Athena, Feature Store
        # - models = Model Registry
        # - endpoints = Sagemaker Endpoints, Model Monitors

        # Pull in AWS Service Connectors
        cls.incoming_data = S3Bucket(cls.incoming_data_bucket)
        cls.data_catalog = DataCatalog(database_scope)
        cls.feature_store = FeatureStore()
        cls.model_registry = ModelRegistry()
        cls.endpoints = Endpoints()
        # Model Monitors

        # Temporal Cache for Metadata
        cls.meta_cache = Cache(10)

        # This connection map sets up the connector objects for each category of metadata
        # Note: Even though this seems confusing, it makes other code WAY simpler
        cls.connection_map = {
            ServiceCategory.INCOMING_DATA: cls.incoming_data,
            ServiceCategory.DATA_CATALOG: cls.data_catalog,
            ServiceCategory.FEATURE_STORE: cls.feature_store,
            ServiceCategory.MODELS: cls.model_registry,
            ServiceCategory.ENDPOINTS: cls.endpoints
        }

    @classmethod
    def refresh_meta(cls, category: ServiceCategory) -> None:
        """Refresh the Metadata for the given Category

        Args:
            category (ServiceCategory): The Category of metadata to Pull
        """
        # Refresh the connection for the given category and pull new data
        cls.connection_map[category].refresh()
        cls.meta_cache.set(category, cls.connection_map[category].metadata())

    @classmethod
    def get_metadata(cls, category: ServiceCategory) -> dict:
        """Pull Metadata for the given Service Category

        Args:
            category (ServiceCategory): The Service Category to pull metadata from

        Returns:
            dict: The Metadata for the Requested Service Category
        """
        # Check the Temporal Cache
        meta_data = cls.meta_cache.get(category)
        if meta_data is not None:
            return meta_data
        else:
            # If we don't have the data in the cache, we need to refresh it
            cls.log.info(f"Refreshing data for {category}...")
            cls.refresh_meta(category)
            return cls.meta_cache.get(category)

    @classmethod
    def get_all_metadata(cls) -> dict:
        """Pull the metadata for ALL the Service Categories"""
        return {_category: cls.get_metadata(_category) for _category in ServiceCategory}


if __name__ == '__main__':
    from pprint import pprint

    # Collect args from the command line
    parser = argparse.ArgumentParser()
    parser.add_argument('--database', type=str, default='sageworks', help='AWS Data Catalog Database')
    args, commands = parser.parse_known_args()

    # Check for unknown args
    if commands:
        print('Unrecognized args: %s' % commands)
        sys.exit(1)

    # Create the class
    aws_broker = AWSServiceBroker(database_scope='all')

    # Get the Metadata for various categories
    for my_category in ServiceCategory:
        print(f"{my_category}:")
        pprint(aws_broker.get_metadata(my_category))

    # Get the Metadata for ALL the categories
    # NOTE: There should be NO Refreshes in the logs
    pprint(aws_broker.get_all_metadata())



"""ModelRegistry: Helper Class for the AWS Model Registry Service"""
import sys
import argparse
from sagemaker.session import Session

# SageWorks Imports
from sageworks.aws_service_broker.aws_service_connectors.connector import Connector


class ModelRegistry(Connector):
    def __init__(self):
        """"ModelRegistry: Helper Class for the AWS Model Registry Service"""
        # Call SuperClass Initialization
        super().__init__()

        # Set up our SageMaker Session and SageMaker Client
        self.sm_session = Session()
        self.sm_client = self.sm_session.boto_session.client("sagemaker")

        # Set up our internal data storage
        self.model_data = {}

    def check(self) -> bool:
        """Check if we can reach/connect to this AWS Service"""
        try:
            self.sm_client.list_model_package_groups()
            return True
        except Exception as e:
            self.log.critical(f"Error connecting to AWS Model Registry Service: {e}")
            return False

    def refresh(self):
        """Load/reload the tables in the database"""
        # Grab all the Model Groups in the AWS Model Registry
        print("Reading Model Registry...")
        _model_groups = self.sm_client.list_model_package_groups()['ModelPackageGroupSummaryList']
        _mg_names = [model_group['ModelPackageGroupName'] for model_group in _model_groups]

        # Get the details for each Model Group and convert to a data structure with direct lookup
        self.model_data = {name: self._model_group_details(name) for name in _mg_names}

    def metadata(self) -> dict:
        """Get all the table information in this database"""
        return self.model_data

    def model_group_names(self) -> list:
        """Get all the feature group names in this database"""
        return list(self.model_data.keys())

    def model_group_details(self, model_group_name: str) -> dict:
        """Get the details for a specific feature group"""
        return self.model_data.get(model_group_name)

    def s3_storage(self, model_group_name: str) -> str:
        """Get the S3 Location for a specific feature group"""
        return 'TBD Later'

    def _model_group_details(self, model_group_name: str) -> dict:
        """Internal: Do not call this method directly, use model_group_details() instead"""

        # Grab the Model Group details from the AWS Model Registry
        details = self.sm_client.list_model_packages(ModelPackageGroupName=model_group_name)['ModelPackageSummaryList']
        for detail in details:
            model_arn = detail['ModelPackageArn']
            detail['ModelPackageDetails'] = self.sm_client.describe_model_package(ModelPackageName=model_arn)
        return details


if __name__ == '__main__':
    from pprint import pprint

    # Collect args from the command line
    parser = argparse.ArgumentParser()
    args, commands = parser.parse_known_args()

    # Check for unknown args
    if commands:
        print('Unrecognized args: %s' % commands)
        sys.exit(1)

    # Create the class and get the AWS Model Registry details
    model_registry = ModelRegistry()
    model_registry.refresh()

    # List the Model Groups
    print('Model Groups:')
    for my_group_name in model_registry.model_group_names():
        print(f"\t{my_group_name}")

    # Get the details for a specific Model Group
    my_group = 'Solubility-Models'
    group_info = model_registry.model_group_details(my_group)
    pprint(group_info)



"""A scatter plot component"""
from dash import dcc
import pandas as pd
import plotly.express as px


# For colormaps see (https://plotly.com/python/discrete-color/#color-sequences-in-plotly-express)
def create(df: pd.DataFrame) -> dcc.Graph:
    """Create a Scatter Plot"""
    gap_df = px.data.gapminder()

    # TEMP
    gap_df.rename({'gdpPercap': 'awesome', 'lifeExp': 'stuff', 'continent': 'Project', 'pop': 'logS'}, axis=1, inplace=True)
    # gap_df.replace({'Asia': 'Project X', 'Europe': 'Project Y', 'Africa': 'Project Z', 'Americas': 'Project A', 'Oceania': 'Project B'}, inplace=True)
    gap_df.replace({'Asia': 'Training Set', 'Europe': 'Training Set', 'Africa': 'Training Set', 'Americas': 'Test Set', 'Oceania': 'Test Set'},
                   inplace=True)

    fig = px.scatter(gap_df,
                     x="awesome", y="stuff", size="logS", color="Project",
                     log_x=True, size_max=60, title="Cool Stuff", color_discrete_sequence=px.colors.qualitative.Plotly)
    return dcc.Graph(id='scatter_plot', figure=fig)



from flask import Flask
app = Flask(__name__)


@app.route("/")
def hello():
    return "<h1 style='color:blue'>Flask Test Application: AOK</h1>"


if __name__ == "__main__":
    app.run(host='0.0.0.0', port=80)



"""PandasToData: Class to publish a Pandas DataFrame as a DataSource"""
import awswrangler as wr
import json
import pandas as pd

# Local imports
from sageworks.utils.logging import logging_setup
from sageworks.transforms.transform import Transform, TransformInput, TransformOutput

# Setup Logging
logging_setup()


class PandasToData(Transform):
    def __init__(self, output_uuid=None):
        """PandasToData: Class to publish a Pandas DataFrame as a DataSource"""

        # Call superclass init
        super().__init__(None, output_uuid)

        # Set up all my instance attributes
        self.input_type = TransformInput.PANDAS_DF
        self.output_type = TransformOutput.DATA_SOURCE
        self.input_df = None

    def transform(self, overwrite: bool = True):
        """Convert the Pandas DataFrame into Parquet Format in the SageWorks S3 Bucket, and
           store the information about the data to the AWS Data Catalog sageworks database"""

        # Add some tags here
        tags = ['sageworks', 'public']

        # Create the Output Parquet file S3 Storage Path
        s3_storage_path = f"{self.data_source_s3_path}/{self.output_uuid}"

        # Write out the DataFrame to Parquet/DataStore/Athena
        wr.s3.to_parquet(self.input_df, path=s3_storage_path, dataset=True, mode='overwrite',
                         database=self.data_catalog_db, table=self.output_uuid,
                         description=f'SageWorks data source: {self.output_uuid}',
                         filename_prefix=f'{self.output_uuid}_',
                         parameters={'tags': json.dumps(tags)},
                         partition_cols=None)  # FIXME: Have some logic around partition columns

    def set_input(self, input_df: pd.DataFrame):
        """Set the DataFrame Input for this Transform"""
        self.input_df = input_df


# Simple test of the PandasToData functionality
def test():
    """Test the PandasToData Class"""
    from datetime import datetime
    from sageworks.artifacts.data_sources.athena_source import AthenaSource

    # Setup Pandas output options
    pd.set_option('display.max_colwidth', 15)
    pd.set_option('display.max_columns', 15)
    pd.set_option('display.width', 1000)

    # Create some fake data
    fake_data = [
        {'id': 1, 'name': 'sue', 'age': 41, 'score': 7.8, 'date': datetime.now()},
        {'id': 2, 'name': 'bob', 'age': 34, 'score': 6.4, 'date': datetime.now()},
        {'id': 3, 'name': 'ted', 'age': 69, 'score': 8.2, 'date': datetime.now()},
        {'id': 4, 'name': 'bill', 'age': 24, 'score': 5.3, 'date': datetime.now()},
        {'id': 5, 'name': 'sally', 'age': 52, 'score': 9.5, 'date': datetime.now()}
        ]
    fake_df = pd.DataFrame(fake_data)

    # Create my DF to Data Source Transform
    output_uuid = 'test_data'
    df_to_data = PandasToData(output_uuid)
    df_to_data.set_input(fake_df)

    # Store this data into a SageWorks DataSource
    df_to_data.transform()

    # Grab the output and query it for a dataframe
    output = AthenaSource(output_uuid)
    df = output.query(f"select * from {output_uuid} limit 5")

    # Show the dataframe
    print(df)


if __name__ == "__main__":
    test()



"""LightTransform: Example Class that demonstrates a Transform for Light DataSources using Pandas"""

# Local imports
from sageworks.transforms.transform import Transform, TransformInput, TransformOutput
from sageworks.transforms.pandas_transforms.data_to_pandas import DataToPandas
from sageworks.transforms.pandas_transforms.pandas_to_data import PandasToData


class LightTransform(Transform):
    def __init__(self, input_uuid: str = None, output_uuid: str = None):
        """"LightTransform: Example Class that demonstrates a Transform for Light DataSources using Pandas"""

        # Call superclass init
        super().__init__(input_uuid, output_uuid)

        # Set up all my instance attributes
        self.input_type = TransformInput.DATA_SOURCE
        self.output_type = TransformOutput.DATA_SOURCE

    def transform(self):
        """Pull the input DataSource, do something kewl, and output to a DataSource"""

        # Grab the Input (Data Source)
        input_df = DataToPandas(self.input_uuid).get_output()  # Shorthand for transform, get_output

        # Do something interesting with the Pandas DataFrame
        interesting_df = input_df  # Add kewl stuff here

        # Now publish to the output location
        output_data = PandasToData(self.output_uuid)
        output_data.set_input(interesting_df)
        output_data.transform()


if __name__ == "__main__":
    """Run/Test the LightTransform Class"""
    from sageworks.artifacts.data_sources.athena_source import AthenaSource

    # Create the class with inputs and outputs and invoke the transform
    my_input = 'aqsol_data'
    my_output = 'aqsol_data_clean'
    LightTransform(my_input, my_output).transform()

    # Use SageWorks to grab the output and query it for a dataframe
    output = AthenaSource(my_output)
    df = output.query(f"select * from {my_output} limit 5")

    # Show the dataframe
    print(df)



"""DFToFeatureSet: Class to publish a Pandas DataFrame into a FeatureSet (Athena/FeatureStore)"""
from datetime import datetime
import logging
import pandas as pd
import time
import botocore
from sagemaker.session import Session
from sagemaker.feature_store.feature_group import FeatureGroup

# Local imports
from sageworks.utils.logging import logging_setup
from sageworks.transforms.transform import Transform, TransformInput, TransformOutput
from sageworks.artifacts.feature_sets.feature_set import FeatureSet
from sageworks.aws_service_broker.aws_sageworks_role_manager import AWSSageWorksRoleManager

# Setup Logging
logging_setup()

# Good Reference for AWS FeatureStore functionality
# - https://sagemaker-examples.readthedocs.io/en/latest/sagemaker-featurestore/feature_store_introduction.html
# AWS ROLES
# IMPORTANT: You must attach the following policies to your execution role: * AmazonS3FullAccess * AmazonSageMakerFeatureStoreAccess


class DFToFeatureSet(Transform):
    def __init__(self):
        """DFToFeatureSet: Class to publish a Pandas DataFrame into a FeatureSet (Athena/FeatureStore)"""

        # Call superclass init
        super().__init__()

        # Set up all my instance attributes
        self.log = logging.getLogger(__name__)
        self.input_df = None
        self.output_uuid = None
        self.data_catalog_db = 'sageworks'
        self.feature_set_s3_path = 's3://sageworks-feature-sets'
        self.id_column = None
        self.event_time_column = None
        self.sagemaker_session = Session()
        self.sageworks_role_arn = AWSSageWorksRoleManager().sageworks_execution_role_arn()

    def input_type(self) -> TransformInput:
        """What Input Type does this Transform Consume"""
        return TransformInput.PANDAS_DF

    def output_type(self) -> TransformOutput:
        """What Output Type does this Transform Produce"""
        return TransformOutput.FEATURE_SET

    def set_input(self, input_df: pd.DataFrame, id_column, event_time_column=None):
        """Set the Input DataFrame for this Transform"""
        self.id_column = id_column
        self.event_time_column = event_time_column
        self.input_df = input_df

    def set_input_uuid(self, input_uuid: str):
        """Not Implemented: Just satisfying the Transform abstract method requirements"""
        pass

    def set_output_uuid(self, output_uuid: str):
        """Set the Name for the output Data Source"""
        self.output_uuid = output_uuid

    def get_output(self) -> FeatureSet:
        """Get the FeatureSet Output from this Transform"""
        return FeatureSet(self.output_uuid)

    def validate_input(self) -> bool:
        """Validate the Input for this Transform"""

        # Simple check on the input dataframe
        return self.input_df is not None and not self.input_df.empty

    def _ensure_event_time(self):
        """Internal: AWS Feature Store requires an EventTime field in all data stored"""
        if self.event_time_column is None or self.event_time_column not in self.input_df.columns:
            current_datetime = datetime.utcnow()
            self.log.info('Generating an EventTime column before FeatureSet Creation...')
            self.event_time_column = 'EventTime'
            self.input_df[self.event_time_column] = pd.Series([current_datetime] * len(self.input_df), dtype="string")
        else:  # The event_time_column is defined so lets make sure it the right type for Feature Store
            if pd.api.types.is_datetime64_dtype(self.input_df[self.event_time_column]):
                self.log.info(f"Converting {self.event_time_column} to ISOFormat Date String before FeatureSet Creation...")

                # Convert the datetime DType to ISO-8601 string
                # Note: We assume that event times are in UTC
                self.input_df[self.event_time_column] = self.input_df[self.event_time_column].map(lambda x: x.isoformat())

    def _convert_objs_to_string(self):
        """Internal: AWS Feature Store doesn't know how to store object dtypes, so convert to String"""
        for col in self.input_df:
            if pd.api.types.is_object_dtype(self.input_df[col].dtype):
                self.input_df[col] = self.input_df[col].astype(pd.StringDtype())

    def transform(self, overwrite: bool = True):
        """Convert the Pandas DataFrame into Parquet Format in the SageWorks S3 Bucket, and
           store the information about the data to the AWS Data Catalog sageworks database"""

        # Ensure that the input dataframe has an EventTime field
        self._ensure_event_time()

        # Convert object dtypes to string
        self._convert_objs_to_string()

        # Are we going to Overwrite the existing Feature Group
        if overwrite:
            try:
                # Delete the Feature Group and ensure that it gets deleted
                remove_fg = FeatureGroup(name=self.output_uuid, sagemaker_session=self.sagemaker_session)
                remove_fg.delete()
                self.ensure_feature_group_deleted(remove_fg)

            except botocore.exceptions.ClientError as error:
                self.log.info(f"Feature Group {self.output_uuid} not found.. this is AOK...")
                self.log.info(error)

        # Create a Feature Group and load our Feature Definitions
        my_feature_group = FeatureGroup(name=self.output_uuid, sagemaker_session=self.sagemaker_session)
        my_feature_group.load_feature_definitions(data_frame=self.input_df)

        # Add some tags here
        tags = ['sageworks', 'public']

        # Create the Output Parquet file S3 Storage Path for this Feature Set
        s3_storage_path = f"{self.feature_set_s3_path}/{self.output_uuid}"

        # Write out the DataFrame to Parquet/FeatureSet/Athena
        my_feature_group.create(
            s3_uri=s3_storage_path,
            record_identifier_name=self.id_column,
            event_time_feature_name=self.event_time_column,
            role_arn=self.sageworks_role_arn,
            enable_online_store=True
        )

        # Ensure/wait for the feature group to be created
        self.ensure_feature_group_created(my_feature_group)

    def ensure_feature_group_created(self, feature_group):
        status = feature_group.describe().get("FeatureGroupStatus")
        while status == "Creating":
            self.log.info("FeatureSet being Created...")
            time.sleep(5)
            status = feature_group.describe().get("FeatureGroupStatus")
        self.log.info(f"FeatureSet {feature_group.name} successfully created")

    def ensure_feature_group_deleted(self, feature_group):
        status = feature_group.describe().get("FeatureGroupStatus")
        while status == "Deleting":
            self.log.info("FeatureSet being Deleted...")
            time.sleep(5)
            status = feature_group.describe().get("FeatureGroupStatus")
        self.log.info(f"FeatureSet {feature_group.name} successfully deleted")


# Simple test of the DFToFeatureSet functionality
def test():
    """Test the DFToFeatureSet Class"""

    # Setup Pandas output options
    pd.set_option('display.max_colwidth', 15)
    pd.set_option('display.max_columns', 15)
    pd.set_option('display.width', 1000)

    # Create some fake data
    fake_data = [
        {'id': 1, 'name': 'sue', 'age': 41, 'score': 7.8, 'date': datetime.utcnow()},
        {'id': 2, 'name': 'bob', 'age': 34, 'score': 6.4, 'date': datetime.utcnow()},
        {'id': 3, 'name': 'ted', 'age': 69, 'score': 8.2, 'date': datetime.utcnow()},
        {'id': 4, 'name': 'bill', 'age': 24, 'score': 5.3, 'date': datetime.utcnow()},
        {'id': 5, 'name': 'sally', 'age': 52, 'score': 9.5, 'date': datetime.utcnow()}
        ]
    fake_df = pd.DataFrame(fake_data)

    # Create my DF to Data Source Transform
    output_uuid = 'test-feature-set'
    df_to_features = DFToFeatureSet()
    df_to_features.set_input(fake_df, id_column='id', event_time_column='date')
    df_to_features.set_output_uuid(output_uuid)

    # Does my data pass validation?
    assert(df_to_features.validate_input())

    # Store this dataframe as a SageWorks Feature Set
    # FIXME: This rest if this test is disabled for now
    """
    df_to_features.transform()

    # Grab the output and query it for a dataframe
    output = df_to_features.get_output()
    query = f'select * from "{output.athena_database}"."{output.athena_table}" limit 5'
    df = output.query(query)
    # Show the dataframe
    print(df)
    """


if __name__ == "__main__":
    test()



"""Class: DataCatalog"""
import sys
import argparse
import awswrangler as wr
import json

# SageWorks Imports
from sageworks.aws_service_broker.aws_service_connectors.connector import Connector


class DataCatalog(Connector):
    """DataCatalog: Helper Class for the AWS Data Catalog"""
    def __init__(self, database_scope: (str, list) = 'sageworks'):
        """DataCatalog: Helper Class for the AWS Data Catalog
           Args: -
               database_scope (str, list) - The scope of the databases to query (default = 'sageworks',
                                            can be list ['a', 'b'], also 'all' for account wide scope)
        """
        # Call SuperClass Initialization
        super().__init__()

        # Set up our database scope
        self.database_scope = database_scope if isinstance(database_scope, list) else [database_scope]
        self.scoped_database_list = self.get_scoped_database_list()

        # Set up our internal data storage
        self.data_catalog_metadata = {}

    def check(self) -> bool:
        """Check if we can reach/connect to this AWS Service"""
        try:
            wr.catalog.get_databases()
            return True
        except Exception as e:
            self.log.critical(f"Error connecting to AWS Data Catalog: {e}")
            return False

    def refresh(self):
        """Load/reload all the tables in all the catalog databases"""

        # For each database in our scoped list, load the tables
        for database in self.scoped_database_list:
            print(f"Reading Data Catalog Database: {database}...")
            table_list = wr.catalog.get_tables(database=database)

            # Convert to a data structure with direct lookup
            self.data_catalog_metadata[database] = {table['Name']: table for table in table_list}

    def metadata(self) -> dict:
        """Get all the table information in this database"""
        return self.data_catalog_metadata

    def get_scoped_database_list(self):
        """Return a list of databases within the defined scope for this class"""
        all_databases = [db['Name'] for db in wr.catalog.get_databases()]
        if self.database_scope == ['all']:
            return all_databases

        # Return a subset of the databases
        return list(set(all_databases).intersection(set(self.database_scope)))

    def get_database_names(self) -> list:
        """Get all the database names from AWS Data Catalog"""
        return list(self.data_catalog_metadata.keys())

    def get_table_names(self, database: str) -> list:
        """Get all the table names in this database"""
        return list(self.data_catalog_metadata[database].keys())

    def get_table(self, database: str, table_name: str) -> dict:
        """Get the table information for the given table name"""
        return self.data_catalog_metadata[database].get(table_name)

    def get_table_tags(self, database: str, table_name: str) -> list:
        """Get the table tag list for the given table name"""
        table = self.get_table(database, table_name)
        return json.loads(table['Parameters'].get('tags', '[]'))

    @staticmethod
    def set_table_tags(database: str, table_name: str, tags: list):
        """Set the tags for a specific table"""
        wr.catalog.upsert_table_parameters(parameters={'tags': json.dumps(tags)},
                                           database=database,
                                           table=table_name)

    @staticmethod
    def add_table_tags(database: str, table_name: str, tags: list):
        """Add some the tags for a specific table"""
        current_tags = json.loads(wr.catalog.get_table_parameters(database, table_name).get('tags'))
        new_tags = list(set(current_tags).union(set(tags)))
        wr.catalog.upsert_table_parameters(parameters={'tags': json.dumps(new_tags)},
                                           database=database,
                                           table=table_name)


if __name__ == '__main__':
    from pprint import pprint

    # Collect args from the command line
    parser = argparse.ArgumentParser()
    args, commands = parser.parse_known_args()

    # Check for unknown args
    if commands:
        print('Unrecognized args: %s' % commands)
        sys.exit(1)

    # Create the class and get the AWS Data Catalog database info
    catalog = DataCatalog()

    # The connectors need an explicit refresh to populate themselves
    catalog.refresh()

    # List databases and tables
    for my_database in catalog.get_database_names():
        print(f"{my_database}")
        for name in catalog.get_table_names(my_database):
            print(f"\t{name}")

    # Get a specific table
    my_database = 'sageworks'
    my_table = 'aqsol_data'
    table_info = catalog.get_table(my_database, my_table)
    pprint(table_info)

    # Get the tags for this table
    my_tags = catalog.get_table_tags(my_database, my_table)
    print(f"Tags: {my_tags}")

    # Set the tags for this table
    catalog.set_table_tags(my_database, my_table, ['public', 'solubility'])

    # Refresh the connector to get the latest info from AWS Data Catalog
    catalog.refresh()

    # Get the tags for this table
    my_tags = catalog.get_table_tags(my_database, my_table)
    print(f"Tags: {my_tags}")

    # Set the tags for this table
    catalog.add_table_tags(my_database, my_table, ['aqsol', 'smiles'])

    # Refresh the connector to get the latest info from AWS Data Catalog
    catalog.refresh()

    # Get the tags for this table
    my_tags = catalog.get_table_tags(my_database, my_table)
    print(f"Tags: {my_tags}")



"""RowTagger: Generalized Row Tagger (domain specific logic should be captured in a subclass"""
# License: Apache 2.0 ©SuperCowPowers LLC
import pandas as pd

# SageWorks Imports
from sageworks.algorithms.table.light import feature_spider


# Class: RowTagger
class RowTagger:
    """RowTagger: Encapsulate business logic and special cases for tagging rows in a dataframe
        - CoIncident (with reasonable difference in target value)
        - High Target Gradient (HTG) Neighborhood
    """

    def __init__(self, dataframe: pd.DataFrame, features: list, id_column: str, target: str,
                 min_dist: float, min_target_diff: float):

        # Set up some parameters
        self.id_column = id_column
        self.min_dist = min_dist
        self.min_target_diff = min_target_diff

        # Do a validation check on the dataframe
        self.df = dataframe
        self.validate_input_data()

        # We need the feature spider for the more advanced tags
        self.f_spider = feature_spider.FeatureSpider(self.df, features, id_column=self.id_column,
                                                     target=target)

        # Add a 'tags' column (if it doesn't already exist)
        if 'tags' not in self.df.columns:
            self.df['tags'] = [set() for _ in range(len(self.df.index))]

    def validate_input_data(self):
        # Make sure it's a dataframe
        if not isinstance(self.df, pd.DataFrame):
            print('Input is NOT a DataFrame!')
            return False

        # Make sure it has some rows and columns
        rows, columns = self.df.shape
        if rows == 0:
            print('Input DataFrame has 0 rows!')
            return False
        if columns == 0:
            print('Input DataFrame has 0 columns!')
            return False

        # Make sure it has an ID column (domain specific stuff in subclass)
        if self.id_column not in self.df.columns:
            print('Input DataFrame needs a ID column!')
            return False

        # AOK
        return True

    def tag_rows(self) -> pd.DataFrame:
        """Run all the current registered taggers"""
        # The taggers that all take file names we want to run
        taggers = [self.coincident, self.high_gradients]
        for tagger in taggers:
            tagger()
        return self.df

    def coincident(self):
        """Find observations with the SAME features that have different target values"""
        coincident_indexes = self.f_spider.coincident(self.min_target_diff, verbose=False)

        # We get back index offsets (not labels) so we need to use iloc
        for index in coincident_indexes:
            self.df['tags'].iloc[index].add('coincident')

    def high_gradients(self):
        """Find observations close in feature space with a high difference in target values
           High Target Gradient (HTG) """
        htg_indexes = self.f_spider.high_gradients(self.min_dist, self.min_target_diff, verbose=False)

        # We get back index offsets (not labels) so we need to use iloc
        for index in htg_indexes:
            self.df['tags'].iloc[index].add('htg')


def test():
    """Test for the RowTagger Class"""

    # Set some pandas options
    pd.set_option('display.max_columns', None)
    pd.set_option('display.width', 1000)

    # Make some fake data
    data = {'ID': ['id_0', 'id_0', 'id_2', 'id_3', 'id_4',
                   'id_5', 'id_6', 'id_7', 'id_8', 'id_9'],
            'feat1': [1.0, 1.0, 1.1, 3.0, 4.0, 1.0, 1.0, 1.1, 3.0, 4.0],
            'feat2': [1.0, 1.0, 1.1, 3.0, 4.0, 1.0, 1.0, 1.1, 3.0, 4.0],
            'feat3': [0.1, 0.1, 0.2, 1.6, 2.5, 0.1, 0.1, 0.2, 1.6, 2.5],
            'price': [31, 60, 62, 40, 20, 31, 61, 60, 40, 20]}
    data_df = pd.DataFrame(data)

    # Create the class and run the taggers
    row_tagger = RowTagger(data_df, ['feat1', 'feat2', 'feat3'], id_column='ID', target='price',
                           min_dist=2.0, min_target_diff=1.0)
    data_df = row_tagger.tag_rows()
    print(data_df)


if __name__ == '__main__':
    test()



"""Data Class for the models"""
import pandas as pd
import json


class ModelData:
    """Data Class for storing and managing the data associated with model"""
    def __init__(self, path):
        self._data = pd.read_csv(path, parse_dates=['date_created'])

    def get_model_df(self) -> pd.DataFrame:
        """Return the underlying Pandas Dataframe"""
        return self._data

    def get_model_details(self, row_index):
        """Get the model details for this model (via row_index)"""
        row = self._data.iloc[row_index]
        return row.to_dict()

    def get_model_confusion_matrix(self, row_index):
        """Get the model confusion matrix for this model (via row_index)"""
        row = self._data.iloc[row_index]
        return json.loads(row['c_matrix'])

    def get_model_feature_importance(self, row_index):
        """Get the model feature importance for this model (via row_index)"""
        row = self._data.iloc[row_index]
        return json.loads(row['feature_importance'])


# Test for our ModelData Class
if __name__ == '__main__':
    models = ModelData('../../data/model_data.csv')
    print(models.get_model_feature_importance(0))
    print(models.get_model_confusion_matrix(1))



"""Layout for the model scoreboard"""
from dash import Dash, html
import dash_bootstrap_components as dbc


def scoreboard_layout(app: Dash, components: dict) -> html.Div:
    layout = html.Div(children=[
            dbc.Row(html.H2(app.title)),
            dbc.Row([
                # Model Table and Model Details
                dbc.Col([
                        dbc.Row(components['model_table']),
                        dbc.Row(html.H3('Model Details'), style={'padding': '50px 0px 0px 0px'}),
                        dbc.Row([dbc.Col(components['model_details']), dbc.Col(components['confusion_matrix'])]),
                        dbc.Row(components['scatter_plot'])
                        ], width=8),

                # Feature Importance and Details
                dbc.Col([
                    dbc.Row(html.H3('Feature Importance')),
                    dbc.Row(components['feature_importance']),
                    dbc.Row(html.H3('Feature Details', style={'padding': '10px 0px 0px 0px'})),
                    dbc.Row(components['feature_details'])
                ], width=4)
            ]),

        ], style={'margin': '30px'})
    return layout


def scoreboard_layout_alt1(app: Dash, components: dict) -> html.Div:
    layout = html.Div(children=[
        dbc.Row(html.H2(app.title)),
        dbc.Row([
            dbc.Col(components['model_table'], width=8),
            dbc.Col(components['confusion_matrix'], width=4)]
        ),
        dbc.Row(components['scatter_plot'])
    ])
    return layout


def scoreboard_layout_alt2(app: Dash, components: dict) -> html.Div:
    layout = html.Div(children=[
        dbc.Row(html.H2(app.title)),
        dbc.Row(html.H2('  ')),
        dbc.Row([
            dbc.Col(components['model_table'], width=7),
            dbc.Col(components['feature_importance'], width=5)]
        ),
        dbc.Row(html.H2('Model Details')),
        dbc.Row([
            dbc.Col(components['model_details'], width=4),
            dbc.Col(components['confusion_matrix'], width=4),
            dbc.Col(components['scatter_plot'], width=4)]
        )
    ])
    return layout



"""AthenaSource: SageWork Data Source accessible through Athena"""
import pandas as pd
import awswrangler as wr
from datetime import datetime

# SageWorks Imports
from sageworks.artifacts.data_sources.data_source import DataSource
from sageworks.aws_service_broker.aws_service_broker import ServiceCategory, AWSServiceBroker


class AthenaSource(DataSource):
    """AthenaSource: SageWork Data Source accessible through Athena"""
    def __init__(self, table_name, database='sageworks'):
        """AthenaSource Initialization

        Args:
            table_name (str): Name of Athena Table
        """

        # Call superclass init
        super().__init__()

        self.data_catalog_db = database
        self.table_name = table_name

        # Grab an AWS Metadata Broker object and pull information for Data Sources
        self.aws_meta = AWSServiceBroker('all')
        self._meta = self.aws_meta.get_metadata(ServiceCategory.DATA_CATALOG)[self.data_catalog_db].get(self.table_name)

        # Grab my tags
        self._tags = self._meta.get('Parameters', {}).get('tags', [])

        # All done
        print(f'AthenaSource Initialized: {self.data_catalog_db}.{self.table_name}')

    def check(self) -> bool:
        """Validation Checks for this Data Source"""

        # We're we able to pull AWS Metadata for this table_name?"""
        if self._meta is None:
            self.log.critical(f'AthenaSource.check() {self.table_name} not found in SageWorks Metadata!')
            return False
        return True

    def deep_check(self) -> bool:
        """These are more comprehensive checks for this Data Source (may take a LONG TIME)"""

        # Can we run an Athena Test Query
        try:
            self.athena_test_query()
            return True
        except Exception as exc:
            self.log.critical(f"Athena Test Query Failed: {exc}")
            return False

    def uuid(self) -> str:
        """The SageWorks Unique Identifier"""
        return self.table_name

    def _aws_uuid(self) -> str:
        """Internal: An AWS Unique Identifier"""
        return f"CatalogID:{self._meta['CatalogId']}"

    def size(self) -> bool:
        """Return the size of this data in MegaBytes"""
        size_in_bytes = sum(wr.s3.size_objects(self.s3_storage_location()).values())
        size_in_mb = round(size_in_bytes / 1_000_000)
        return size_in_mb

    def meta(self):
        """Get the metadata for this artifact"""
        return self._meta

    def tags(self):
        """Get the tags for this artifact"""
        return self._tags

    def add_tag(self, tag):
        """Add a tag to this artifact"""
        # This ensures no duplicate tags
        self._tags = list(set(self._tags).union([tag]))

    def aws_url(self):
        """The AWS URL for looking at/querying this data source"""
        return 'https://us-west-2.console.aws.amazon.com/athena/home'

    def created(self) -> datetime:
        """Return the datetime when this artifact was created"""
        return self._meta['CreateTime']

    def modified(self) -> datetime:
        """Return the datetime when this artifact was last modified"""
        return self._meta['UpdateTime']

    def num_rows(self) -> int:
        """Return the number of rows for this Data Source"""
        count_df = self.query(f'select count(*) AS count from "{self.data_catalog_db}"."{self.table_name}"')
        return count_df['count'][0]

    def num_columns(self) -> int:
        """Return the number of columns for this Data Source"""
        return len(self.column_names())

    def column_names(self) -> list[str]:
        """Return the column names for this Athena Table"""
        return [item['Name'] for item in self._meta['StorageDescriptor']['Columns']]

    def query(self, query: str) -> pd.DataFrame:
        """Query the AthenaSource"""
        df = wr.athena.read_sql_query(sql=query, database=self.data_catalog_db)
        scanned_bytes = df.query_metadata["Statistics"]["DataScannedInBytes"]
        self.log.info(f"Athena Query successful (scanned bytes: {scanned_bytes})")
        return df

    def s3_storage_location(self) -> str:
        """Get the S3 Storage Location for this Data Source"""
        return self._meta['StorageDescriptor']['Location']

    def athena_test_query(self):
        """Validate that Athena Queries are working"""
        query = f"select count(*) as count from {self.table_name}"
        df = wr.athena.read_sql_query(sql=query, database=self.data_catalog_db)
        scanned_bytes = df.query_metadata["Statistics"]["DataScannedInBytes"]
        self.log.info(f"Athena TEST Query successful (scanned bytes: {scanned_bytes})")


# Simple test of the AthenaSource functionality
def test():
    """Test for AthenaSource Class"""

    # Retrieve a Data Source
    my_data = AthenaSource('aqsol_data')

    # Call the various methods

    # Verify that the Athena Data Source exists
    assert(my_data.check())

    # What's my SageWorks UUID
    print(f"UUID: {my_data.uuid()}")

    # Get the S3 Storage for this Data Source
    print(f"S3 Storage: {my_data.s3_storage_location()}")

    # What's the size of the data?
    print(f"Size of Data (MB): {my_data.size()}")

    # When was it created and last modified?
    print(f"Created: {my_data.created()}")
    print(f"Modified: {my_data.modified()}")

    # How many rows and columns?
    num_rows = my_data.num_rows()
    num_columns = my_data.num_columns()
    print(f'Rows: {num_rows} Columns: {num_columns}')

    # What are the column names?
    columns = my_data.column_names()
    print(columns)

    # Get all the metadata associated with this data source
    print(f"Meta: {my_data.meta()}")

    # Get the tags associated with this data source
    print(f"Tags: {my_data.tags()}")

    # Run a query to only pull back a few columns and rows
    column_query = ', '.join(columns[:3])
    query = f'select {column_query} from "{my_data.data_catalog_db}"."{my_data.table_name}" limit 10'
    df = my_data.query(query)
    print(df)


if __name__ == "__main__":
    test()



"""Show a list of files/tables and tags from the given AWS catalog database"""
import sys
import argparse
from os.path import basename

# Local imports
from glue_utils.glue_meta import GlueMeta


# Class: List Tags
class ListTags:
    """ListTags: """
    def __init__(self, catalog_database: str):
        # Our Glue Meta Class takes care of the low level details
        self.glue_meta = GlueMeta(catalog_database, skip_ignore=False)

    def list_tags(self):
        """List all the files/tags for the tables in this AWS catalog database"""
        catalog_table_data = self.glue_meta.get_table_data()
        for name, table_info in catalog_table_data.items():
            file_name = basename(table_info['location'])
            print(f"{file_name}   Tags: {table_info['tags']}...")
            # print(f"File: {table_info['location']}...")


if __name__ == '__main__':

    # Collect args from the command line
    parser = argparse.ArgumentParser()
    parser.add_argument('--database', type=str, default='sageworks', help='The AWS catalog database name')
    args, commands = parser.parse_known_args()

    # Check for unknown args
    if commands:
        print('Unrecognized args: %s' % commands)
        sys.exit(1)

    # Create our Class and List the Tags
    list_tags = ListTags(args.database)
    list_tags.list_tags()



"""Setup logging defaults"""
import logging
import os


def logging_setup():
    """This will setup the default formatters and handlers for logging"""
    logging.basicConfig(stream=os.sys.stdout,
                        level=os.environ.get("LOGLEVEL", "INFO"),
                        format='%(asctime)s (%(filename)s:%(lineno)d) %(levelname)s %(message)s',
                        datefmt='%Y-%m-%d %H:%M:%S')

    # Make boto be more quiet
    logging.getLogger('boto3').setLevel(logging.CRITICAL)
    logging.getLogger('botocore').setLevel(logging.CRITICAL)



"""Utility/helper methods for Pandas dataframe operations"""
import pandas as pd


def get_percent_null(df):
    print('Dataframe ({:d} rows)'.format(len(df)))
    s = df.isna().mean().round(3) * 100.0
    s.name = '%Null'
    return s


def unique(df):
    s = df.nunique()
    s.name = '#unique'
    return s


def examples(df):
    first_n = [df[c].unique()[:10].tolist() for c in df.columns]
    s = pd.Series(first_n, df.columns)
    s.name = 'Examples'
    return s


def info(df):
    s1 = get_percent_null(df)
    s2 = unique(df)
    s3 = examples(df)
    df = pd.concat([s1, s2, s3], axis=1)
    return df.sort_values('%Null')



"""DataSource: Abstract Base Class for all data sources (S3: CSV, Parquet, RDS, etc)"""
from abc import abstractmethod
import pandas as pd
import awswrangler as wr

# Local imports
from sageworks.artifacts.artifact import Artifact


class DataSource(Artifact):
    def __init__(self):
        """DataSource: Abstract Base Class for all data sources (S3: CSV, Parquet, RDS, etc)"""

        # Call superclass init
        super().__init__()

        # Make sure the AWS data catalog database exists
        self.ensure_aws_catalog_db()

    @abstractmethod
    def num_rows(self) -> int:
        """Return the number of rows for this Data Source"""
        pass

    @abstractmethod
    def num_columns(self) -> int:
        """Return the number of columns for this Data Source"""
        pass

    @abstractmethod
    def column_names(self) -> list[str]:
        """Return the column names for this Data Source"""
        pass

    @abstractmethod
    def query(self, query: str) -> pd.DataFrame:
        """Query the DataSource"""
        pass

    def ensure_aws_catalog_db(self):
        """Ensure that the AWS Catalog Database exists"""
        wr.catalog.create_database(self.data_catalog_db, exist_ok=True)



"""FeatureSet: SageWork Feature Set accessible through Athena"""
import logging

# SageWorks Imports
from sageworks.artifacts.data_sources.athena_source import AthenaSource
from sageworks.aws_service_broker.aws_service_broker import ServiceCategory, AWSServiceBroker
from sageworks.utils.logging import logging_setup

# Setup Logging
logging_setup()


class FeatureSet(AthenaSource):

    def __init__(self, feature_set_name):
        """FeatureSet: SageWork Feature Set accessible through Athena

        Args:
            feature_set_name (str): Name of Feature Set in SageWorks Metadata.
        """
        self.log = logging.getLogger(__name__)
        self.feature_set_name = feature_set_name

        # Grab an AWS Metadata Broker object and pull information for Feature Sets
        self.aws_meta = AWSServiceBroker('all')
        self.feature_set_meta = self.aws_meta.get_metadata(ServiceCategory.FEATURE_STORE).get(self.feature_set_name)

        # Pull Athena and S3 Storage information from metadata
        self.athena_database = self.feature_set_meta['sageworks'].get('athena_database')
        self.athena_table = self.feature_set_meta['sageworks'].get('athena_table')
        self.s3_storage = self.feature_set_meta['sageworks'].get('s3_storage')

        # Call SuperClass (AthenaSource) Initialization
        super().__init__(self.athena_table, self.athena_database)

        # All done
        print(f'FeatureSet Initialized: {feature_set_name}')

    def check(self) -> bool:
        """Does the feature_set_name exist in the AWS Metadata?"""
        if self.feature_set_meta is None:
            self.log.critical(f'FeatureSet.check() {self.feature_set_name} not found in AWS Metadata!')
            return False
        return True


# Simple test of the FeatureSet functionality
def test():
    """Test for FeatureSet Class"""

    # Grab a FeatureSet object and pull some information from it
    my_features = FeatureSet('AqSolDB-base')

    # Call the various methods

    # Lets do a check/validation of the feature set
    assert(my_features.check())

    # How many rows and columns?
    num_rows = my_features.num_rows()
    num_columns = my_features.num_columns()
    print(f'Rows: {num_rows} Columns: {num_columns}')

    # What are the column names?
    columns = my_features.column_names()
    print(columns)

    # Get the tags associated with this feature set
    print(f"Tags: {my_features.tags()}")

    # Run a query to only pull back a few columns and rows
    column_query = ', '.join(columns[:3])
    query = f'select {column_query} from "{my_features.athena_database}"."{my_features.athena_table}" limit 10'
    df = my_features.query(query)
    print(df)


if __name__ == "__main__":
    test()



"""DataToPandas: Class to transform a Data Source into a Pandas DataFrame"""
import pandas as pd

# Local imports
from sageworks.transforms.transform import Transform, TransformInput, TransformOutput
from sageworks.artifacts.data_sources.athena_source import AthenaSource


class DataToPandas(Transform):
    def __init__(self, input_uuid=None):
        """DataToPandas: Class to transform a Data Source into a Pandas DataFrame"""

        # Call superclass init
        super().__init__(input_uuid, None)

        # Set up all my instance attributes
        self.input_type = TransformInput.DATA_SOURCE
        self.output_type = TransformOutput.PANDAS_DF
        self.output_df = None
        self.transform_run = False

    def transform(self, max_rows=100000):
        """Convert the DataSource into a Pandas DataFrame"""

        # Grab the Input (Data Source)
        input_data = AthenaSource(self.input_uuid)
        if not input_data.check():
            self.log.critical(f"Data Check on {self.input_uuid} failed!")
            return

        # Get the number of rows in the DataSource
        num_rows = input_data.num_rows()

        # If the data source has more rows than max_rows, do a sample query
        if num_rows > max_rows:
            percentage = round(max_rows*100.0/num_rows)
            self.log.warning(f"DataSource has {num_rows} rows.. sampling down to {max_rows}...")
            query = f"SELECT * FROM {self.input_uuid} TABLESAMPLE BERNOULLI({percentage})"
        else:
            query = f"SELECT * FROM {self.input_uuid}"

        # Mark the transform as complete and set the output DataFrame
        self.transform_run = True
        self.output_df = input_data.query(query)

    def get_output(self) -> pd.DataFrame:
        """Get the DataFrame Output from this Transform"""
        if not self.transform_run:
            self.transform()
        return self.output_df


# Simple test of the DataToPandas functionality
def test():
    """Test the DataToPandas Class"""

    # Setup Pandas output options
    pd.set_option('display.max_colwidth', 15)
    pd.set_option('display.max_columns', 15)
    pd.set_option('display.width', 1000)

    # Grab a Data Source
    data_uuid = 'aqsol_data'

    # Create the DataSource to DF Transform
    data_to_df = DataToPandas(data_uuid)

    # Transform the DataSource into a Pandas DataFrame (with max_rows = 1000)
    data_to_df.transform(max_rows=1000)

    # Grab the output dataframe and show it
    my_df = data_to_df.get_output()
    print(my_df)


if __name__ == "__main__":
    test()



"""Feature Importance component"""
from dash import dcc
import pandas as pd
import plotly.express as px


def create_figure(feature_importance: dict, orientation='h') -> dcc.Graph:
    """Create the Figure for a Feature Importance Chart"""
    df = pd.DataFrame(
        {'Feature': feature_importance.keys(),
         'Importance': feature_importance.values()},
    )
    df = df.sort_values(by='Importance')
    fig = px.bar(df, x='Importance', y='Feature', orientation=orientation)
    fig.update_layout(
        margin=dict(l=20, r=20, t=20, b=20),
    )
    return fig


def create(feature_importance: dict, orientation='h') -> dcc.Graph:
    """Create a Feature Importance Chart"""
    return dcc.Graph(id='feature_importance', figure=create_figure(feature_importance, orientation))



"""AWSSageWorksRoleManager provides a bit of logic/functionality over the set of AWS IAM Services"""
import sys
import boto3
from botocore.exceptions import ClientError
import argparse
import logging

# SageWorks Imports
from sageworks.utils.logging import logging_setup

# Setup Logging
logging_setup()


class AWSSageWorksRoleManager:

    def __init__(self, role_name='SageWorks-ExecutionRole'):
        """"AWSSageWorksRoleManagerSession: Get the SageWorks Execution Role and/or Session"""
        self.log = logging.getLogger(__file__)
        self.role_name = role_name

    @staticmethod
    def check_aws_identity() -> bool:
        """Check the AWS Identity currently active"""
        # Check AWS Identity Token
        sts = boto3.client('sts')
        try:
            identity = sts.get_caller_identity()
            print("\nAWS Account Info:")
            print(f"\tAccount: {identity['Account']}")
            print(f"\tARN: {identity['Arn']}")
            return True
        except ClientError:
            print("AWS Identity Check Failure: Check AWS_PROFILE and Renew Security Token...")
            return False

    def sageworks_execution_role_arn(self):
        """Get the SageWorks Execution Role"""
        try:
            iam = boto3.client('iam')
            role_arn = iam.get_role(RoleName=self.role_name)['Role']['Arn']
            return role_arn
        except iam.exceptions.NoSuchEntityException:
            print(f"Could Not Find Role {self.role_name}")
            return None

    def sageworks_session(self):
        """Create a sageworks session using sts.assume_role(sageworks_execution_role)"""
        # Make sure we can access stuff with this role
        session = boto3.Session()

        sts = session.client("sts")
        response = sts.assume_role(
            RoleArn=self.sageworks_execution_role_arn(),
            RoleSessionName="sageworks-session"
        )
        new_session = boto3.Session(aws_access_key_id=response['Credentials']['AccessKeyId'],
                                    aws_secret_access_key=response['Credentials']['SecretAccessKey'],
                                    aws_session_token=response['Credentials']['SessionToken'])
        return new_session


if __name__ == '__main__':

    # Collect args from the command line
    parser = argparse.ArgumentParser()
    args, commands = parser.parse_known_args()

    # Check for unknown args
    if commands:
        print('Unrecognized args: %s' % commands)
        sys.exit(1)

    # Create the class
    sageworks_role = AWSSageWorksRoleManager()

    # Check our AWS identity
    sageworks_role.check_aws_identity()

    # Get our Execution Role ARN
    role_arn = sageworks_role.sageworks_execution_role_arn()
    print(role_arn)

    # Get our SageWorks Session
    sageworks_session = sageworks_role.sageworks_session()
    print(sageworks_session)

    # Try to access some AWS services
    s3 = sageworks_session.client("s3")
    print(s3.list_buckets())



#!/usr/bin/env python
"""Setup.py for SageWorks: Sagemaker Workbench"""

import os
import glob

from setuptools import setup, find_packages

readme = open('README.md').read()


# Data and Example Files
def get_files(dir_name):
    """Simple directory walker"""
    return [(os.path.join('.', d), [os.path.join(d, f) for f in files]) for d, _, files in os.walk(dir_name)]


setup(
    name='sageworks',
    # use_scm_version=True,
    version='0.1.1',
    description='SageWorks: An easy to use WorkBench for creating and deploying SageMaker Models',
    long_description=readme,
    long_description_content_type='text/markdown',
    author='SuperCowPowers LLC',
    author_email='support@supercowpowers.com',
    url='https://github.com/SuperCowPowers/sageworks',
    packages=find_packages('src'),
    package_dir={"": "src"},
    py_modules=[os.path.splitext(os.path.basename(path))[0] for path in glob.glob("src/*.py")],
    include_package_data=True,
    data_files=get_files('data') + get_files('examples'),
    install_requires=[
        'boto3',
        'awswrangler',
        'sagemaker',
        'pandas',
        'scikit-learn',
        'dash'
    ],
    license='Apache License 2.0',
    keywords='SageMaker, Machine Learning, AWS, Python, Utilities',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy'
    ],
    setup_requires=["setuptools_scm", "setuptools"]
)



"""Transform: Base Class for all transforms within SageWorks
              Inherited Classes must implement the abstract transform() method"""
from abc import ABC, abstractmethod
from enum import Enum, auto
import logging

from sageworks.utils.logging import logging_setup

# Setup Logging
logging_setup()


class TransformInput(Enum):
    """Enumerated Types for SageWorks Transform Inputs"""
    PANDAS_DF = auto()
    SPARK_DF = auto()
    S3_OBJECT = auto()
    DATA_SOURCE = auto()
    FEATURE_SET = auto()
    MODEL = auto()


class TransformOutput(Enum):
    """Enumerated Types for SageWorks Transform Outputs"""
    PANDAS_DF = auto()
    SPARK_DF = auto()
    S3_OBJECT = auto()
    DATA_SOURCE = auto()
    FEATURE_SET = auto()
    MODEL = auto()
    ENDPOINT = auto()


class Transform(ABC):
    def __init__(self, input_uuid: str = None, output_uuid: str = None):
        """Transform: Abstract Base Class for all transforms in SageWorks"""

        # FIXME: Should this class be a Python dataclass?
        self.log = logging.getLogger(__name__)
        self.input_type = None
        self.output_type = None
        self.input_uuid = input_uuid
        self.output_uuid = output_uuid

        # FIXME: We should have this come from AWS or Config
        self.data_catalog_db = 'sageworks'
        self.data_source_s3_path = 's3://sageworks-data-sources'
        self.feature_sets_s3_path = 's3://sageworks-feature-sets'

    @abstractmethod
    def transform(self, overwrite: bool = True):
        """Perform the Transformation from Input to Output
           Args:
               overwrite (bool): Overwrite the output_uuid if it exists (default = True)
        """
        pass

    def input_type(self) -> TransformInput:
        """What Input Type does this Transform Consume"""
        return self.input_type

    def output_type(self) -> TransformOutput:
        """What Output Type does this Transform Produce"""
        return self.output_type

    def set_input_uuid(self, input_uuid: str):
        """Set the Input UUID (Name) for this Transform"""
        self.input_uuid = input_uuid

    def set_output_uuid(self, output_uuid: str):
        """Set the Output UUID (Name) for this Transform"""
        self.output_uuid = output_uuid
